@@ -2,21 +2,28 @@
 Command line interface for working with flow runs
 """
 
+from __future__ import annotations
+
 import logging
 import os
+import signal
+import threading
+import webbrowser
+from types import FrameType
 from typing import List, Optional
 from uuid import UUID
 
 import httpx
-import pendulum
+import orjson
 import typer
-from prefect._vendor.starlette import status
+from rich.markup import escape
 from rich.pretty import Pretty
 from rich.table import Table
+from starlette import status
 
 from prefect.cli._types import PrefectTyper
 from prefect.cli._utilities import exit_with_error, exit_with_success
-from prefect.cli.root import app
+from prefect.cli.root import app, is_interactive
 from prefect.client.orchestration import get_client
 from prefect.client.schemas.filters import FlowFilter, FlowRunFilter, LogFilter
 from prefect.client.schemas.objects import StateType
@@ -26,23 +33,42 @@ from prefect.exceptions import ObjectNotFound
 from prefect.logging import get_logger
 from prefect.runner import Runner
 from prefect.states import State
+from prefect.types._datetime import human_friendly_diff
+from prefect.utilities.asyncutils import run_sync_in_worker_thread
+from prefect.utilities.urls import url_for
 
-flow_run_app = PrefectTyper(
-    name="flow-run", help="Commands for interacting with flow runs."
+flow_run_app: PrefectTyper = PrefectTyper(
+    name="flow-run", help="Interact with flow runs."
 )
 app.add_typer(flow_run_app, aliases=["flow-runs"])
 
 LOGS_DEFAULT_PAGE_SIZE = 200
 LOGS_WITH_LIMIT_FLAG_DEFAULT_NUM_LOGS = 20
 
-logger = get_logger(__name__)
+logger: "logging.Logger" = get_logger(__name__)
 
 
 @flow_run_app.command()
-async def inspect(id: UUID):
+async def inspect(
+    id: UUID,
+    web: bool = typer.Option(
+        False,
+        "--web",
+        help="Open the flow run in a web browser.",
+    ),
+    output: Optional[str] = typer.Option(
+        None,
+        "--output",
+        "-o",
+        help="Specify an output format. Currently supports: json",
+    ),
+):
     """
     View details about a flow run.
     """
+    if output and output.lower() != "json":
+        exit_with_error("Only 'json' output format is supported.")
+
     async with get_client() as client:
         try:
             flow_run = await client.read_flow_run(id)
@@ -52,7 +78,24 @@ async def inspect(id: UUID):
             else:
                 raise
 
-    app.console.print(Pretty(flow_run))
+    if web:
+        flow_run_url = url_for("flow-run", obj_id=id)
+        if not flow_run_url:
+            exit_with_error(
+                "Failed to generate URL for flow run. Make sure PREFECT_UI_URL is configured."
+            )
+
+        await run_sync_in_worker_thread(webbrowser.open_new_tab, flow_run_url)
+        exit_with_success(f"Opened flow run {id!r} in browser.")
+    else:
+        if output and output.lower() == "json":
+            flow_run_json = flow_run.model_dump(mode="json")
+            json_output = orjson.dumps(
+                flow_run_json, option=orjson.OPT_INDENT_2
+            ).decode()
+            app.console.print(json_output)
+        else:
+            app.console.print(Pretty(flow_run))
 
 
 @flow_run_app.command()
@@ -61,6 +104,12 @@ async def ls(
     limit: int = typer.Option(15, help="Maximum number of flow runs to list"),
     state: List[str] = typer.Option(None, help="Name of the flow run's state"),
     state_type: List[str] = typer.Option(None, help="Type of the flow run's state"),
+    output: Optional[str] = typer.Option(
+        None,
+        "--output",
+        "-o",
+        help="Specify an output format. Currently supports: json",
+    ),
 ):
     """
     View recent flow runs or flow runs for specific flows.
@@ -85,6 +134,8 @@ async def ls(
 
     $ prefect flow-runs ls --state-type RUNNING --state-type FAILED
     """
+    if output and output.lower() != "json":
+        exit_with_error("Only 'json' output format is supported.")
 
     # Handling `state` and `state_type` argument validity in the function instead of by specifying
     # List[StateType] and List[StateName] in the type hints, allows users to provide
@@ -150,31 +201,39 @@ async def ls(
         }
 
         if not flow_runs:
+            if output and output.lower() == "json":
+                app.console.print("[]")
+                return
             exit_with_success("No flow runs found.")
 
-    table = Table(title="Flow Runs")
-    table.add_column("ID", justify="right", style="cyan", no_wrap=True)
-    table.add_column("Flow", style="blue", no_wrap=True)
-    table.add_column("Name", style="green", no_wrap=True)
-    table.add_column("State", no_wrap=True)
-    table.add_column("When", style="bold", no_wrap=True)
+    if output and output.lower() == "json":
+        flow_runs_json = [flow_run.model_dump(mode="json") for flow_run in flow_runs]
+        json_output = orjson.dumps(flow_runs_json, option=orjson.OPT_INDENT_2).decode()
+        app.console.print(json_output)
+    else:
+        table = Table(title="Flow Runs")
+        table.add_column("ID", justify="right", style="cyan", no_wrap=True)
+        table.add_column("Flow", style="blue", no_wrap=True)
+        table.add_column("Name", style="green", no_wrap=True)
+        table.add_column("State", no_wrap=True)
+        table.add_column("When", style="bold", no_wrap=True)
 
-    for flow_run in sorted(flow_runs, key=lambda d: d.created, reverse=True):
-        flow = flows_by_id[flow_run.flow_id]
-        timestamp = (
-            flow_run.state.state_details.scheduled_time
-            if flow_run.state.is_scheduled()
-            else flow_run.state.timestamp
-        )
-        table.add_row(
-            str(flow_run.id),
-            str(flow.name),
-            str(flow_run.name),
-            str(flow_run.state.type.value),
-            pendulum.instance(timestamp).diff_for_humans(),
-        )
+        for flow_run in sorted(flow_runs, key=lambda d: d.created, reverse=True):
+            flow = flows_by_id[flow_run.flow_id]
+            timestamp = (
+                flow_run.state.state_details.scheduled_time
+                if flow_run.state.is_scheduled()
+                else flow_run.state.timestamp
+            )
+            table.add_row(
+                str(flow_run.id),
+                str(flow.name),
+                str(flow_run.name),
+                str(flow_run.state.type.value),
+                human_friendly_diff(timestamp),
+            )
 
-    app.console.print(table)
+        app.console.print(table)
 
 
 @flow_run_app.command()
@@ -184,6 +243,11 @@ async def delete(id: UUID):
     """
     async with get_client() as client:
         try:
+            if is_interactive() and not typer.confirm(
+                (f"Are you sure you want to delete flow run with id {id!r}?"),
+                default=False,
+            ):
+                exit_with_error("Deletion aborted.")
             await client.delete_flow_run(id)
         except ObjectNotFound:
             exit_with_error(f"Flow run '{id}' not found!")
@@ -301,13 +365,14 @@ async def logs(
             )
 
             for log in reversed(page_logs) if tail and not reverse else page_logs:
+                # Print following the flow run format (declared in logging.yml)
+                timestamp = f"{log.timestamp:%Y-%m-%d %H:%M:%S.%f}"[:-3]
+                log_level = f"{logging.getLevelName(log.level):7s}"
+                flow_run_info = f"Flow run {flow_run.name!r} - {escape(log.message)}"
+
+                log_message = f"{timestamp} | {log_level} | {flow_run_info}"
                 app.console.print(
-                    # Print following the flow run format (declared in logging.yml)
-                    (
-                        f"{pendulum.instance(log.timestamp).to_datetime_string()}.{log.timestamp.microsecond // 1000:03d} |"
-                        f" {logging.getLevelName(log.level):7s} | Flow run"
-                        f" {flow_run.name!r} - {log.message}"
-                    ),
+                    log_message,
                     soft_wrap=True,
                 )
 
@@ -345,4 +410,19 @@ async def execute(
     if id is None:
         exit_with_error("Could not determine the ID of the flow run to execute.")
 
-    await Runner().execute_flow_run(id)
+    runner = Runner()
+
+    def _handle_reschedule_sigterm(_signal: int, _frame: FrameType | None):
+        logger.info("SIGTERM received, initiating graceful shutdown...")
+        runner.reschedule_current_flow_runs()
+        exit_with_success("Flow run successfully rescheduled.")
+
+    # Set up signal handling to reschedule run on SIGTERM
+    on_sigterm = os.environ.get("PREFECT_FLOW_RUN_EXECUTE_SIGTERM_BEHAVIOR", "").lower()
+    if (
+        threading.current_thread() is threading.main_thread()
+        and on_sigterm == "reschedule"
+    ):
+        signal.signal(signal.SIGTERM, _handle_reschedule_sigterm)
+
+    await runner.execute_flow_run(id)
