@@ -1,18 +1,16 @@
-from enum import Enum
 from pathlib import Path
 from typing import Optional
 
 import orjson
-import pendulum
 import typer
+from pydantic import ValidationError
 from rich.pretty import Pretty
 from rich.table import Table
 
-from prefect import get_client
-from prefect._internal.pydantic import HAS_PYDANTIC_V2
 from prefect.cli._types import PrefectTyper
 from prefect.cli._utilities import exit_with_error, exit_with_success
-from prefect.cli.root import app
+from prefect.cli.root import app, is_interactive
+from prefect.client.orchestration import get_client
 from prefect.client.schemas.actions import (
     GlobalConcurrencyLimitCreate,
     GlobalConcurrencyLimitUpdate,
@@ -21,24 +19,16 @@ from prefect.exceptions import (
     ObjectNotFound,
     PrefectHTTPStatusError,
 )
+from prefect.types._datetime import human_friendly_diff
 
-if HAS_PYDANTIC_V2:
-    from pydantic.v1 import ValidationError
-else:
-    from pydantic import ValidationError
-
-global_concurrency_limit_app = PrefectTyper(
+global_concurrency_limit_app: PrefectTyper = PrefectTyper(
     name="global-concurrency-limit",
-    help="Commands for managing global concurrency limits.",
+    help="Manage global concurrency limits.",
 )
 
 app.add_typer(global_concurrency_limit_app, aliases=["gcl"])
 
 
-class OutputFormat(Enum):
-    JSON = "json"
-
-
 @global_concurrency_limit_app.command("ls")
 async def list_global_concurrency_limits():
     """
@@ -65,6 +55,8 @@ async def list_global_concurrency_limits():
     table.add_column("Updated", style="blue", no_wrap=True)
 
     for gcl_limit in sorted(gcl_limits, key=lambda x: f"{x.name}"):
+        assert gcl_limit.created is not None, "created is not None"
+        assert gcl_limit.updated is not None, "updated is not None"
         table.add_row(
             str(gcl_limit.id),
             gcl_limit.name,
@@ -72,8 +64,8 @@ async def list_global_concurrency_limits():
             str(gcl_limit.limit),
             str(gcl_limit.active_slots),
             str(gcl_limit.slot_decay_per_second),
-            pendulum.instance(gcl_limit.created).isoformat(),
-            pendulum.instance(gcl_limit.updated).diff_for_humans(),
+            gcl_limit.created.isoformat(),
+            human_friendly_diff(gcl_limit.updated),
         )
 
     app.console.print(table)
@@ -84,12 +76,11 @@ async def inspect_global_concurrency_limit(
     name: str = typer.Argument(
         ..., help="The name of the global concurrency limit to inspect."
     ),
-    output: Optional[OutputFormat] = typer.Option(
+    output: Optional[str] = typer.Option(
         None,
         "--output",
         "-o",
-        help="Output format for the command.",
-        case_sensitive=False,
+        help="Specify an output format. Currently supports: json",
     ),
     file_path: Optional[Path] = typer.Option(
         None,
@@ -117,6 +108,9 @@ async def inspect_global_concurrency_limit(
         slot_decay_per_second (float): The slot decay per second.
 
     """
+    if output and output.lower() != "json":
+        exit_with_error("Only 'json' output format is supported.")
+
     if file_path and not output:
         exit_with_error("The --file/-f option requires the --output option to be set.")
 
@@ -126,19 +120,17 @@ async def inspect_global_concurrency_limit(
         except ObjectNotFound:
             exit_with_error(f"Global concurrency limit {name!r} not found.")
 
-    if output:
-        gcl_limit = gcl_limit.dict(json_compatible=True)
-        json_output = orjson.dumps(gcl_limit, option=orjson.OPT_INDENT_2).decode()
+    if output and output.lower() == "json":
+        gcl_limit_json = gcl_limit.model_dump(mode="json")
+        json_output = orjson.dumps(gcl_limit_json, option=orjson.OPT_INDENT_2).decode()
         if not file_path:
             app.console.print(json_output)
-
         else:
             with open(file_path, "w") as f:
                 f.write(json_output)
                 exit_with_success(
                     f"Global concurrency limit {name!r} written to {file_path}"
                 )
-
     else:
         app.console.print(Pretty(gcl_limit))
 
@@ -155,14 +147,16 @@ async def delete_global_concurrency_limit(
     Arguments:
         name (str): The name of the global concurrency limit to delete.
     """
-    if not typer.confirm(
-        f"Are you sure you want to delete global concurrency limit {name!r}?",
-        default=False,
-    ):
-        exit_with_error("Deletion aborted.")
-
     async with get_client() as client:
         try:
+            gcl_limit = await client.read_global_concurrency_limit_by_name(name=name)
+
+            if is_interactive() and not typer.confirm(
+                f"Are you sure you want to delete global concurrency limit with name {gcl_limit.name!r}?",
+                default=False,
+            ):
+                exit_with_error("Deletion aborted.")
+
             await client.delete_global_concurrency_limit_by_name(name=name)
         except ObjectNotFound:
             exit_with_error(f"Global concurrency limit {name!r} not found.")
@@ -288,11 +282,11 @@ async def update_global_concurrency_limit(
     if slot_decay_per_second is not None:
         gcl.slot_decay_per_second = slot_decay_per_second
 
-    if not gcl.dict(exclude_unset=True, shallow=True):
+    if not gcl.model_dump(exclude_unset=True):
         exit_with_error("No update arguments provided.")
 
     try:
-        GlobalConcurrencyLimitUpdate(**gcl.dict())
+        GlobalConcurrencyLimitUpdate(**gcl.model_dump())
     except ValidationError as exc:
         exit_with_error(f"Invalid arguments provided: {exc}")
     except Exception as exc:
