@@ -1,27 +1,31 @@
+from __future__ import annotations
+
 import shutil
 import subprocess
 from copy import deepcopy
 from pathlib import Path
-from typing import Any, Dict, Optional, Protocol, TypedDict, Union, runtime_checkable
+from typing import (
+    Any,
+    Optional,
+    Protocol,
+    TypedDict,
+    Union,
+    runtime_checkable,
+)
 from urllib.parse import urlparse, urlsplit, urlunparse
 from uuid import uuid4
 
-import fsspec
+import fsspec  # pyright: ignore[reportMissingTypeStubs]
 from anyio import run_process
+from pydantic import SecretStr
 
 from prefect._internal.concurrency.api import create_call, from_async
-from prefect._internal.pydantic import HAS_PYDANTIC_V2
 from prefect.blocks.core import Block, BlockNotSavedError
 from prefect.blocks.system import Secret
 from prefect.filesystems import ReadableDeploymentStorage, WritableDeploymentStorage
 from prefect.logging.loggers import get_logger
 from prefect.utilities.collections import visit_collection
 
-if HAS_PYDANTIC_V2:
-    from pydantic.v1 import SecretStr
-else:
-    from pydantic import SecretStr
-
 
 @runtime_checkable
 class RunnerStorage(Protocol):
@@ -30,7 +34,7 @@ class RunnerStorage(Protocol):
     remotely stored flow code.
     """
 
-    def set_base_path(self, path: Path):
+    def set_base_path(self, path: Path) -> None:
         """
         Sets the base path to use when pulling contents from remote storage to
         local storage.
@@ -52,20 +56,20 @@ class RunnerStorage(Protocol):
         """
         ...
 
-    async def pull_code(self):
+    async def pull_code(self) -> None:
         """
         Pulls contents from remote storage to the local filesystem.
         """
         ...
 
-    def to_pull_step(self) -> dict:
+    def to_pull_step(self) -> dict[str, Any] | list[dict[str, Any]]:
         """
         Returns a dictionary representation of the storage object that can be
         used as a deployment pull step.
         """
         ...
 
-    def __eq__(self, __value) -> bool:
+    def __eq__(self, __value: Any) -> bool:
         """
         Equality check for runner storage objects.
         """
@@ -74,7 +78,40 @@ class RunnerStorage(Protocol):
 
 class GitCredentials(TypedDict, total=False):
     username: str
-    access_token: Union[str, Secret]
+    access_token: str | Secret[str]
+
+
+@runtime_checkable
+class _GitCredentialsFormatter(Protocol):
+    """
+    Protocol for credential blocks that can format themselves for git URLs.
+
+    Implementing this protocol allows credential blocks to own their auth
+    formatting logic instead of having it centralized in core Prefect.
+
+    This enables proper separation of concerns where each git provider
+    (GitLab, GitHub, BitBucket) can handle their own authentication format
+    requirements without core needing provider-specific knowledge.
+    """
+
+    def format_git_credentials(self, url: str) -> str:
+        """
+        Format and return the full git URL with credentials embedded.
+
+        Args:
+            url: The repository URL (e.g., "https://gitlab.com/org/repo.git").
+
+        Returns:
+            Complete URL with credentials embedded in the appropriate format for the provider.
+
+        Examples:
+            - GitLab PAT: "https://oauth2:my-token@gitlab.com/org/repo.git"
+            - GitLab Deploy Token: "https://username:token@gitlab.com/org/repo.git"
+            - GitHub: "https://my-token@github.com/org/repo.git"
+            - BitBucket Cloud: "https://x-token-auth:my-token@bitbucket.org/org/repo.git"
+            - BitBucket Server: "https://username:token@bitbucketserver.com/scm/project/repo.git"
+        """
+        ...
 
 
 class GitRepository:
@@ -92,6 +129,7 @@ class GitRepository:
         pull_interval: The interval in seconds at which to pull contents from
             remote storage to local storage. If None, remote storage will perform
             a one-time sync.
+        directories: The directories to pull from the Git repository (uses git sparse-checkout)
 
     Examples:
         Pull the contents of a private git repository to the local filesystem:
@@ -111,11 +149,13 @@ class GitRepository:
     def __init__(
         self,
         url: str,
-        credentials: Union[GitCredentials, Block, Dict[str, Any], None] = None,
-        name: Optional[str] = None,
-        branch: Optional[str] = None,
+        credentials: Union[GitCredentials, Block, dict[str, Any], None] = None,
+        name: str | None = None,
+        branch: str | None = None,
+        commit_sha: str | None = None,
         include_submodules: bool = False,
-        pull_interval: Optional[int] = 60,
+        pull_interval: int | None = 60,
+        directories: list[str] | None = None,
     ):
         if credentials is None:
             credentials = {}
@@ -129,22 +169,31 @@ class GitRepository:
                 "If a username is provided, an access token or password must also be"
                 " provided."
             )
+
+        if branch and commit_sha:
+            raise ValueError(
+                "Cannot provide both a branch and a commit SHA. Please provide only one."
+            )
+
         self._url = url
         self._branch = branch
+        self._commit_sha = commit_sha
         self._credentials = credentials
         self._include_submodules = include_submodules
         repo_name = urlparse(url).path.split("/")[-1].replace(".git", "")
-        default_name = f"{repo_name}-{branch}" if branch else repo_name
+        safe_branch = branch.replace("/", "-") if branch else None
+        default_name = f"{repo_name}-{safe_branch}" if safe_branch else repo_name
         self._name = name or default_name
         self._logger = get_logger(f"runner.storage.git-repository.{self._name}")
         self._storage_base_path = Path.cwd()
         self._pull_interval = pull_interval
+        self._directories = directories
 
     @property
     def destination(self) -> Path:
         return self._storage_base_path / self._name
 
-    def set_base_path(self, path: Path):
+    def set_base_path(self, path: Path) -> None:
         self._storage_base_path = path
 
     @property
@@ -153,13 +202,17 @@ class GitRepository:
 
     @property
     def _repository_url_with_credentials(self) -> str:
+        """Get the repository URL with credentials embedded."""
         if not self._credentials:
             return self._url
 
-        url_components = urlparse(self._url)
+        # If block implements the protocol, let it format the complete URL
+        if isinstance(self._credentials, _GitCredentialsFormatter):
+            return self._credentials.format_git_credentials(self._url)
 
+        # Otherwise, use legacy formatting for plain dict credentials
         credentials = (
-            self._credentials.dict()
+            self._credentials.model_dump()
             if isinstance(self._credentials, Block)
             else deepcopy(self._credentials)
         )
@@ -170,20 +223,109 @@ class GitRepository:
             elif isinstance(v, SecretStr):
                 credentials[k] = v.get_secret_value()
 
-        formatted_credentials = _format_token_from_credentials(
-            urlparse(self._url).netloc, credentials
+        # Get credential string for plain dict credentials
+        block = self._credentials if isinstance(self._credentials, Block) else None
+        credential_string = _format_token_from_credentials(
+            urlparse(self._url).netloc, credentials, block
         )
-        if url_components.scheme == "https" and formatted_credentials is not None:
-            updated_components = url_components._replace(
-                netloc=f"{formatted_credentials}@{url_components.netloc}"
+
+        # Insert credentials into URL
+        components = urlparse(self._url)
+        if components.scheme != "https":
+            return self._url
+
+        return urlunparse(
+            components._replace(netloc=f"{credential_string}@{components.netloc}")
+        )
+
+    @property
+    def _git_config(self) -> list[str]:
+        """Build a git configuration to use when running git commands."""
+        config: dict[str, str] = {}
+
+        # Submodules can be private. The url in .gitmodules
+        # will not include the credentials, we need to
+        # propagate them down here if they exist.
+        if self._include_submodules and self._credentials:
+            # Get base URL (without path) with credentials
+            base_url_parsed = urlparse(self._url)._replace(path="")
+            base_url_without_auth = urlunparse(base_url_parsed)
+
+            # Create a temporary URL with just the base to get credentials formatting
+            if isinstance(self._credentials, _GitCredentialsFormatter):
+                base_url_with_auth = self._credentials.format_git_credentials(
+                    base_url_without_auth
+                )
+            else:
+                # Use legacy credential insertion
+                credentials_dict = (
+                    self._credentials.model_dump()
+                    if isinstance(self._credentials, Block)
+                    else deepcopy(self._credentials)
+                )
+                for k, v in credentials_dict.items():
+                    if isinstance(v, Secret):
+                        credentials_dict[k] = v.get()
+                    elif isinstance(v, SecretStr):
+                        credentials_dict[k] = v.get_secret_value()
+
+                block = (
+                    self._credentials if isinstance(self._credentials, Block) else None
+                )
+                credential_string = _format_token_from_credentials(
+                    base_url_parsed.netloc, credentials_dict, block
+                )
+                base_url_with_auth = urlunparse(
+                    base_url_parsed._replace(
+                        netloc=f"{credential_string}@{base_url_parsed.netloc}"
+                    )
+                )
+
+            config[f"url.{base_url_with_auth}.insteadOf"] = base_url_without_auth
+
+        return ["-c", " ".join(f"{k}={v}" for k, v in config.items())] if config else []
+
+    async def is_sparsely_checked_out(self) -> bool:
+        """
+        Check if existing repo is sparsely checked out
+        """
+        try:
+            result = await run_process(
+                ["git", "config", "--get", "core.sparseCheckout"], cwd=self.destination
             )
-            repository_url = urlunparse(updated_components)
-        else:
-            repository_url = self._url
+            return result.stdout.decode().strip().lower() == "true"
+        except Exception:
+            return False
 
-        return repository_url
+    async def is_shallow_clone(self) -> bool:
+        """
+        Check if the repository is a shallow clone
+        """
+        try:
+            result = await run_process(
+                ["git", "rev-parse", "--is-shallow-repository"],
+                cwd=self.destination,
+            )
+            return result.stdout.decode().strip().lower() == "true"
+        except Exception:
+            return False
 
-    async def pull_code(self):
+    async def is_current_commit(self) -> bool:
+        """
+        Check if the current commit is the same as the commit SHA
+        """
+        if not self._commit_sha:
+            raise ValueError("No commit SHA provided")
+        try:
+            result = await run_process(
+                ["git", "rev-parse", self._commit_sha],
+                cwd=self.destination,
+            )
+            return result.stdout.decode().strip() == self._commit_sha
+        except Exception:
+            return False
+
+    async def pull_code(self) -> None:
         """
         Pulls the contents of the configured repository to the local filesystem.
         """
@@ -202,8 +344,7 @@ class GitRepository:
                 cwd=str(self.destination),
             )
             existing_repo_url = None
-            if result.stdout is not None:
-                existing_repo_url = _strip_auth_from_url(result.stdout.decode().strip())
+            existing_repo_url = _strip_auth_from_url(result.stdout.decode().strip())
 
             if existing_repo_url != self._url:
                 raise ValueError(
@@ -211,23 +352,64 @@ class GitRepository:
                     f"does not match the configured repository {self._url}"
                 )
 
+            # Sparsely checkout the repository if directories are specified and the repo is not in sparse-checkout mode already
+            if self._directories and not await self.is_sparsely_checked_out():
+                await run_process(
+                    ["git", "sparse-checkout", "set", *self._directories],
+                    cwd=self.destination,
+                )
+
             self._logger.debug("Pulling latest changes from origin/%s", self._branch)
             # Update the existing repository
-            cmd = ["git", "pull", "origin"]
-            if self._branch:
-                cmd += [self._branch]
-            if self._include_submodules:
-                cmd += ["--recurse-submodules"]
-            cmd += ["--depth", "1"]
-            try:
-                await run_process(cmd, cwd=self.destination)
-                self._logger.debug("Successfully pulled latest changes")
-            except subprocess.CalledProcessError as exc:
-                self._logger.error(
-                    f"Failed to pull latest changes with exit code {exc}"
+            cmd = ["git"]
+            # Add the git configuration, must be given after `git` and before the command
+            cmd += self._git_config
+
+            # If the commit is already checked out, skip the pull
+            if self._commit_sha and await self.is_current_commit():
+                return
+
+            # If checking out a specific commit, fetch the latest changes and unshallow the repository if necessary
+            elif self._commit_sha:
+                if await self.is_shallow_clone():
+                    cmd += ["fetch", "origin", "--unshallow"]
+                else:
+                    cmd += ["fetch", "origin", self._commit_sha]
+                try:
+                    await run_process(cmd, cwd=self.destination)
+                    self._logger.debug("Successfully fetched latest changes")
+                except subprocess.CalledProcessError as exc:
+                    self._logger.error(
+                        f"Failed to fetch latest changes with exit code {exc}"
+                    )
+                    shutil.rmtree(self.destination)
+                    await self._clone_repo()
+
+                await run_process(
+                    ["git", "checkout", self._commit_sha],
+                    cwd=self.destination,
                 )
-                shutil.rmtree(self.destination)
-                await self._clone_repo()
+                self._logger.debug(
+                    f"Successfully checked out commit {self._commit_sha}"
+                )
+
+            # Otherwise, pull the latest changes from the branch
+            else:
+                cmd += ["pull", "origin"]
+                if self._branch:
+                    cmd += [self._branch]
+                if self._include_submodules:
+                    cmd += ["--recurse-submodules"]
+                cmd += ["--depth", "1"]
+                try:
+                    await run_process(cmd, cwd=self.destination)
+                    self._logger.debug("Successfully pulled latest changes")
+                except subprocess.CalledProcessError as exc:
+                    self._logger.error(
+                        f"Failed to pull latest changes with exit code {exc}"
+                    )
+                    shutil.rmtree(self.destination)
+                    await self._clone_repo()
 
         else:
             await self._clone_repo()
@@ -239,31 +421,69 @@ class GitRepository:
         self._logger.debug("Cloning repository %s", self._url)
 
         repository_url = self._repository_url_with_credentials
+        cmd = ["git"]
+        # Add the git configuration, must be given after `git` and before the command
+        cmd += self._git_config
+        # Add the clone command and its parameters
+        cmd += ["clone", repository_url]
 
-        cmd = [
-            "git",
-            "clone",
-            repository_url,
-        ]
-        if self._branch:
-            cmd += ["--branch", self._branch]
         if self._include_submodules:
             cmd += ["--recurse-submodules"]
 
-        # Limit git history and set path to clone to
-        cmd += ["--depth", "1", str(self.destination)]
+        # This will only checkout the top-level directory
+        if self._directories:
+            cmd += ["--sparse"]
+
+        if self._commit_sha:
+            cmd += ["--filter=blob:none", "--no-checkout"]
+
+        else:
+            if self._branch:
+                cmd += ["--branch", self._branch]
+
+            # Limit git history
+            cmd += ["--depth", "1"]
+
+        # Set path to clone to
+        cmd += [str(self.destination)]
 
         try:
             await run_process(cmd)
         except subprocess.CalledProcessError as exc:
             # Hide the command used to avoid leaking the access token
-            exc_chain = None if self._credentials else exc
+            parsed_url = urlparse(self._url)
+            exc_chain = (
+                None
+                if self._credentials or parsed_url.password or parsed_url.username
+                else exc
+            )
             raise RuntimeError(
-                f"Failed to clone repository {self._url!r} with exit code"
+                f"Failed to clone repository {_strip_auth_from_url(self._url)!r} with exit code"
                 f" {exc.returncode}."
             ) from exc_chain
 
-    def __eq__(self, __value) -> bool:
+        if self._commit_sha:
+            # Fetch the commit
+            await run_process(
+                ["git", "fetch", "origin", self._commit_sha],
+                cwd=self.destination,
+            )
+            # Checkout the specific commit
+            await run_process(
+                ["git", "checkout", self._commit_sha],
+                cwd=self.destination,
+            )
+            self._logger.debug(f"Successfully checked out commit {self._commit_sha}")
+
+        # Once repository is cloned and the repo is in sparse-checkout mode then grow the working directory
+        if self._directories:
+            self._logger.debug("Will add %s", self._directories)
+            await run_process(
+                ["git", "sparse-checkout", "set", *self._directories],
+                cwd=self.destination,
+            )
+
+    def __eq__(self, __value: Any) -> bool:
         if isinstance(__value, GitRepository):
             return (
                 self._url == __value._url
@@ -278,28 +498,33 @@ class GitRepository:
             f" branch={self._branch!r})"
         )
 
-    def to_pull_step(self) -> Dict:
-        pull_step = {
+    def to_pull_step(self) -> dict[str, Any]:
+        pull_step: dict[str, Any] = {
             "prefect.deployments.steps.git_clone": {
                 "repository": self._url,
                 "branch": self._branch,
             }
         }
         if self._include_submodules:
-            pull_step["prefect.deployments.steps.git_clone"][
-                "include_submodules"
-            ] = self._include_submodules
+            pull_step["prefect.deployments.steps.git_clone"]["include_submodules"] = (
+                self._include_submodules
+            )
+        if self._commit_sha:
+            pull_step["prefect.deployments.steps.git_clone"]["commit_sha"] = (
+                self._commit_sha
+            )
         if isinstance(self._credentials, Block):
-            pull_step["prefect.deployments.steps.git_clone"][
-                "credentials"
-            ] = f"{{{{ {self._credentials.get_block_placeholder()} }}}}"
-        elif isinstance(self._credentials, dict):
-            if isinstance(self._credentials.get("access_token"), Secret):
+            pull_step["prefect.deployments.steps.git_clone"]["credentials"] = (
+                f"{{{{ {self._credentials.get_block_placeholder()} }}}}"
+            )
+        elif isinstance(self._credentials, dict):  # pyright: ignore[reportUnnecessaryIsInstance]
+            if isinstance(
+                access_token := self._credentials.get("access_token"), Secret
+            ):
                 pull_step["prefect.deployments.steps.git_clone"]["credentials"] = {
                     **self._credentials,
                     "access_token": (
-                        "{{"
-                        f" {self._credentials['access_token'].get_block_placeholder()} }}}}"
+                        f"{{{{ {access_token.get_block_placeholder()} }}}}"
                     ),
                 }
             elif self._credentials.get("access_token") is not None:
@@ -391,21 +616,21 @@ class RemoteStorage:
 
         def replace_blocks_with_values(obj: Any) -> Any:
             if isinstance(obj, Block):
-                if hasattr(obj, "get"):
-                    return obj.get()
+                if get := getattr(obj, "get", None):
+                    return get()
                 if hasattr(obj, "value"):
-                    return obj.value
+                    return getattr(obj, "value")
                 else:
-                    return obj.dict()
+                    return obj.model_dump()
             return obj
 
         settings_with_block_values = visit_collection(
             self._settings, replace_blocks_with_values, return_data=True
         )
 
-        return fsspec.filesystem(scheme, **settings_with_block_values)
+        return fsspec.filesystem(scheme, **settings_with_block_values)  # pyright: ignore[reportUnknownMemberType] missing type stubs
 
-    def set_base_path(self, path: Path):
+    def set_base_path(self, path: Path) -> None:
         self._storage_base_path = path
 
     @property
@@ -431,7 +656,7 @@ class RemoteStorage:
         _, netloc, urlpath, _, _ = urlsplit(self._url)
         return Path(netloc) / Path(urlpath.lstrip("/"))
 
-    async def pull_code(self):
+    async def pull_code(self) -> None:
         """
         Pulls contents from remote storage to the local filesystem.
         """
@@ -449,7 +674,7 @@ class RemoteStorage:
         try:
             await from_async.wait_for_call_in_new_thread(
                 create_call(
-                    self._filesystem.get,
+                    self._filesystem.get,  # pyright: ignore[reportUnknownArgumentType, reportUnknownMemberType] missing type stubs
                     remote_path,
                     str(self.destination),
                     recursive=True,
@@ -461,7 +686,7 @@ class RemoteStorage:
                 f" {self.destination!r}"
             ) from exc
 
-    def to_pull_step(self) -> dict:
+    def to_pull_step(self) -> dict[str, Any]:
         """
         Returns a dictionary representation of the storage object that can be
         used as a deployment pull step.
@@ -485,12 +710,12 @@ class RemoteStorage:
             }
         }
         if required_package:
-            step["prefect.deployments.steps.pull_from_remote_storage"][
-                "requires"
-            ] = required_package
+            step["prefect.deployments.steps.pull_from_remote_storage"]["requires"] = (
+                required_package
+            )
         return step
 
-    def __eq__(self, __value) -> bool:
+    def __eq__(self, __value: Any) -> bool:
         """
         Equality check for runner storage objects.
         """
@@ -516,20 +741,16 @@ class BlockStorageAdapter:
         self._block = block
         self._pull_interval = pull_interval
         self._storage_base_path = Path.cwd()
-        if not isinstance(block, Block):
+        if not isinstance(block, Block):  # pyright: ignore[reportUnnecessaryIsInstance]
             raise TypeError(
                 f"Expected a block object. Received a {type(block).__name__!r} object."
             )
         if not hasattr(block, "get_directory"):
             raise ValueError("Provided block must have a `get_directory` method.")
 
-        self._name = (
-            f"{block.get_block_type_slug()}-{block._block_document_name}"
-            if block._block_document_name
-            else str(uuid4())
-        )
+        self._name = f"{block.get_block_type_slug()}-{getattr(block, '_block_document_name', uuid4())}"
 
-    def set_base_path(self, path: Path):
+    def set_base_path(self, path: Path) -> None:
         self._storage_base_path = path
 
     @property
@@ -540,17 +761,17 @@ class BlockStorageAdapter:
     def destination(self) -> Path:
         return self._storage_base_path / self._name
 
-    async def pull_code(self):
+    async def pull_code(self) -> None:
         if not self.destination.exists():
             self.destination.mkdir(parents=True, exist_ok=True)
         await self._block.get_directory(local_path=str(self.destination))
 
-    def to_pull_step(self) -> dict:
-        # Give blocks the change to implement their own pull step
+    def to_pull_step(self) -> dict[str, Any]:
+        # Give blocks the chance to implement their own pull step
         if hasattr(self._block, "get_pull_step"):
-            return self._block.get_pull_step()
+            return getattr(self._block, "get_pull_step")()
         else:
-            if not self._block._block_document_name:
+            if getattr(self._block, "_block_document_name", None) is None:
                 raise BlockNotSavedError(
                     "Block must be saved with `.save()` before it can be converted to a"
                     " pull step."
@@ -558,11 +779,11 @@ class BlockStorageAdapter:
             return {
                 "prefect.deployments.steps.pull_with_block": {
                     "block_type_slug": self._block.get_block_type_slug(),
-                    "block_document_name": self._block._block_document_name,
+                    "block_document_name": getattr(self._block, "_block_document_name"),
                 }
             }
 
-    def __eq__(self, __value) -> bool:
+    def __eq__(self, __value: Any) -> bool:
         if isinstance(__value, BlockStorageAdapter):
             return self._block == __value._block
         return False
@@ -571,16 +792,12 @@ class BlockStorageAdapter:
 class LocalStorage:
     """
     Sets the working directory in the local filesystem.
-
     Parameters:
         Path: Local file path to set the working directory for the flow
-
     Examples:
         Sets the working directory for the local path to the flow:
-
         ```python
         from prefect.runner.storage import Localstorage
-
         storage = LocalStorage(
             path="/path/to/local/flow_directory",
         )
@@ -601,19 +818,19 @@ class LocalStorage:
     def destination(self) -> Path:
         return self._path
 
-    def set_base_path(self, path: Path):
+    def set_base_path(self, path: Path) -> None:
         self._storage_base_path = path
 
     @property
     def pull_interval(self) -> Optional[int]:
         return self._pull_interval
 
-    async def pull_code(self):
+    async def pull_code(self) -> None:
         # Local storage assumes the code already exists on the local filesystem
         # and does not need to be pulled from a remote location
         pass
 
-    def to_pull_step(self) -> dict:
+    def to_pull_step(self) -> dict[str, Any]:
         """
         Returns a dictionary representation of the storage object that can be
         used as a deployment pull step.
@@ -625,7 +842,7 @@ class LocalStorage:
         }
         return step
 
-    def __eq__(self, __value) -> bool:
+    def __eq__(self, __value: Any) -> bool:
         if isinstance(__value, LocalStorage):
             return self._path == __value._path
         return False
@@ -663,20 +880,35 @@ def create_storage_from_source(
         return LocalStorage(path=source, pull_interval=pull_interval)
 
 
-def _format_token_from_credentials(netloc: str, credentials: dict) -> str:
+def _format_token_from_credentials(
+    netloc: str,
+    credentials: dict[str, Any] | GitCredentials,
+    block: Block | None = None,
+) -> str:
     """
     Formats the credentials block for the git provider.
 
-    BitBucket supports the following syntax:
-        git clone "https://x-token-auth:{token}@bitbucket.org/yourRepoOwnerHere/RepoNameHere"
-        git clone https://username:<token>@bitbucketserver.com/scm/projectname/teamsinspace.git
+    If the block implements _GitCredentialsFormatter protocol, delegates to it.
+    Otherwise, uses generic formatting for plain dict credentials.
+
+    Args:
+        netloc: The network location (hostname) of the git repository
+        credentials: Dictionary containing credential information
+        block: Optional Block object that may implement the formatter protocol
     """
+    if block is not None and isinstance(block, _GitCredentialsFormatter):
+        # Reconstruct full URL for context (scheme doesn't matter for formatting)
+        url = f"https://{netloc}"
+        return block.format_git_credentials(url)
     username = credentials.get("username") if credentials else None
     password = credentials.get("password") if credentials else None
     token = credentials.get("token") if credentials else None
     access_token = credentials.get("access_token") if credentials else None
 
-    user_provided_token = access_token or token or password
+    user_provided_token: str | Secret[str] | None = access_token or token or password
+
+    if isinstance(user_provided_token, Secret):
+        user_provided_token = user_provided_token.get()
 
     if not user_provided_token:
         raise ValueError(
@@ -687,38 +919,7 @@ def _format_token_from_credentials(netloc: str, credentials: dict) -> str:
     if username:
         return f"{username}:{user_provided_token}"
 
-    if "bitbucketserver" in netloc:
-        # If they pass a BitBucketCredentials block and we don't have both a username and at
-        # least one of a password or token and they don't provide a header themselves,
-        # we can raise the appropriate error to avoid the wrong format for BitBucket Server.
-        if not username and ":" not in user_provided_token:
-            raise ValueError(
-                "Please provide a `username` and a `password` or `token` in your"
-                " BitBucketCredentials block to clone a repo from BitBucket Server."
-            )
-        # if username or if no username but it's provided in the token
-        return (
-            f"{username}:{user_provided_token}"
-            if username and username not in user_provided_token
-            else user_provided_token
-        )
-
-    elif "bitbucket" in netloc:
-        return (
-            user_provided_token
-            if user_provided_token.startswith("x-token-auth:")
-            or ":" in user_provided_token
-            else f"x-token-auth:{user_provided_token}"
-        )
-
-    elif "gitlab" in netloc:
-        return (
-            f"oauth2:{user_provided_token}"
-            if not user_provided_token.startswith("oauth2:")
-            else user_provided_token
-        )
-
-    # all other cases (GitHub, etc.)
+    # Fallback for plain dict credentials without a block
     return user_provided_token
 
 
@@ -727,7 +928,7 @@ def _strip_auth_from_url(url: str) -> str:
 
     # Construct a new netloc without the auth info
     netloc = parsed.hostname
-    if parsed.port:
+    if parsed.port and netloc:
         netloc += f":{parsed.port}"
 
     # Build the sanitized URL
