@@ -7,25 +7,26 @@ from typing import List, Optional
 from uuid import UUID
 
 import jsonschema.exceptions
-import pendulum
-from prefect._vendor.fastapi import Body, Depends, HTTPException, Path, Response, status
-from prefect._vendor.starlette.background import BackgroundTasks
+import sqlalchemy as sa
+from fastapi import Body, Depends, HTTPException, Path, Response
 
 import prefect.server.api.dependencies as dependencies
 import prefect.server.models as models
 import prefect.server.schemas as schemas
+from prefect._internal.compatibility.starlette import status
 from prefect.server.api.validation import (
     validate_job_variables_for_deployment,
     validate_job_variables_for_deployment_flow_run,
 )
 from prefect.server.api.workers import WorkerLookups
-from prefect.server.database.dependencies import provide_database_interface
-from prefect.server.database.interface import PrefectDBInterface
+from prefect.server.database import PrefectDBInterface, provide_database_interface
 from prefect.server.exceptions import MissingVariableError, ObjectNotFoundError
 from prefect.server.models.deployments import mark_deployments_ready
 from prefect.server.models.workers import DEFAULT_AGENT_WORK_POOL_NAME
-from prefect.server.utilities.schemas import DateTimeTZ
+from prefect.server.schemas.responses import DeploymentPaginationResponse
 from prefect.server.utilities.server import PrefectRouter
+from prefect.types import DateTime
+from prefect.types._datetime import now
 from prefect.utilities.schema_tools.hydration import (
     HydrationContext,
     HydrationError,
@@ -37,7 +38,7 @@ from prefect.utilities.schema_tools.validation import (
     validate,
 )
 
-router = PrefectRouter(prefix="/deployments", tags=["Deployments"])
+router: PrefectRouter = PrefectRouter(prefix="/deployments", tags=["Deployments"])
 
 
 def _multiple_schedules_error(deployment_id) -> HTTPException:
@@ -62,16 +63,18 @@ async def create_deployment(
     db: PrefectDBInterface = Depends(provide_database_interface),
 ) -> schemas.responses.DeploymentResponse:
     """
-    Gracefully creates a new deployment from the provided schema. If a deployment with
+    Creates a new deployment from the provided schema. If a deployment with
     the same name and flow_id already exists, the deployment is updated.
 
     If the deployment has an active schedule, flow runs will be scheduled.
     When upserting, any scheduled runs from the existing deployment will be deleted.
+
+    For more information, see https://docs.prefect.io/v3/concepts/deployments.
     """
 
-    data = deployment.dict(exclude_unset=True)
-    data["created_by"] = created_by.dict() if created_by else None
-    data["updated_by"] = updated_by.dict() if created_by else None
+    data = deployment.model_dump(exclude_unset=True)
+    data["created_by"] = created_by.model_dump() if created_by else None
+    data["updated_by"] = updated_by.model_dump() if created_by else None
 
     async with db.session_context(begin_transaction=True) as session:
         if (
@@ -95,7 +98,31 @@ async def create_deployment(
             )
 
         # hydrate the input model into a full model
-        deployment_dict = deployment.dict(exclude={"work_pool_name"})
+        deployment_dict: dict = deployment.model_dump(
+            exclude={"work_pool_name"},
+            exclude_unset=True,
+        )
+
+        requested_concurrency_limit = deployment_dict.pop(
+            "global_concurrency_limit_id", "unset"
+        )
+        if requested_concurrency_limit != "unset":
+            if requested_concurrency_limit:
+                concurrency_limit = (
+                    await models.concurrency_limits_v2.read_concurrency_limit(
+                        session=session,
+                        concurrency_limit_id=requested_concurrency_limit,
+                    )
+                )
+
+                if not concurrency_limit:
+                    raise HTTPException(
+                        status_code=status.HTTP_404_NOT_FOUND,
+                        detail="Concurrency limit not found",
+                    )
+
+            deployment_dict["concurrency_limit_id"] = requested_concurrency_limit
+
         if deployment.work_pool_name and deployment.work_queue_name:
             # If a specific pool name/queue name combination was provided, get the
             # ID for that work pool queue.
@@ -161,29 +188,25 @@ async def create_deployment(
                     ),
                 )
 
-        # Ensure that `paused` and `is_schedule_active` are consistent.
-        if "paused" in data:
-            deployment.is_schedule_active = not data["paused"]
-        elif "is_schedule_active" in data:
-            deployment.paused = not data["is_schedule_active"]
-
-        now = pendulum.now("UTC")
+        right_now = now("UTC")
         model = await models.deployments.create_deployment(
             session=session, deployment=deployment
         )
 
-        if model.created >= now:
+        if model.created >= right_now:
             response.status_code = status.HTTP_201_CREATED
 
-        return schemas.responses.DeploymentResponse.from_orm(model)
+        return schemas.responses.DeploymentResponse.model_validate(
+            model, from_attributes=True
+        )
 
 
-@router.patch("/{id}", status_code=status.HTTP_204_NO_CONTENT)
+@router.patch("/{id:uuid}", status_code=status.HTTP_204_NO_CONTENT)
 async def update_deployment(
     deployment: schemas.actions.DeploymentUpdate,
     deployment_id: UUID = Path(..., description="The deployment id", alias="id"),
     db: PrefectDBInterface = Depends(provide_database_interface),
-):
+) -> None:
     async with db.session_context(begin_transaction=True) as session:
         existing_deployment = await models.deployments.read_deployment(
             session=session, deployment_id=deployment_id
@@ -193,43 +216,40 @@ async def update_deployment(
                 status.HTTP_404_NOT_FOUND, detail="Deployment not found."
             )
 
-        # We need to make sure that when a user PATCHes a deployment with
-        # `schedule` and `is_schedule_active` fields, that we populate the
-        # `schedules` field so that the one-to-many relationship is updated.
-        # This is support for legacy clients that don't know about the
-        # `schedules` field.
+        # Checking how we should handle schedule updates
+        # If not all existing schedules have slugs then we'll fall back to the existing logic where are schedules are recreated to match the request.
+        # If the existing schedules have slugs, but not all provided schedules have slugs, then we'll return a 422 to avoid accidentally blowing away schedules.
+        # Otherwise, we'll use the existing slugs and the provided slugs to make targeted updates to the deployment's schedules.
+        schedules_to_patch: list[schemas.actions.DeploymentScheduleUpdate] = []
+        schedules_to_create: list[schemas.actions.DeploymentScheduleUpdate] = []
+        all_provided_have_slugs = all(
+            schedule.slug is not None for schedule in deployment.schedules or []
+        )
+        all_existing_have_slugs = existing_deployment.schedules and all(
+            schedule.slug is not None for schedule in existing_deployment.schedules
+        )
+        if all_provided_have_slugs and all_existing_have_slugs:
+            current_slugs = [
+                schedule.slug for schedule in existing_deployment.schedules
+            ]
 
-        update_data = deployment.dict(exclude_unset=True)
-
-        if "schedule" in update_data or "is_schedule_active" in update_data:
-            if len(existing_deployment.schedules) > 1:
-                raise _multiple_schedules_error(deployment_id)
-
-            schedule = (
-                update_data["schedule"]
-                if "schedule" in update_data
-                else existing_deployment.schedule
-            )
-            active = (
-                update_data["is_schedule_active"]
-                if "is_schedule_active" in update_data
-                else existing_deployment.is_schedule_active
-            )
-
-            if schedule is not None:
-                deployment.schedules = [
-                    schemas.actions.DeploymentScheduleCreate(
-                        schedule=schedule, active=active
+            for schedule in deployment.schedules:
+                if schedule.slug in current_slugs:
+                    schedules_to_patch.append(schedule)
+                elif schedule.schedule:
+                    schedules_to_create.append(schedule)
+                else:
+                    raise HTTPException(
+                        status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
+                        detail="Unable to create new deployment schedules without a schedule configuration.",
                     )
-                ]
-            elif schedule is None:
-                deployment.schedules = []
-
-        # Ensure that `paused` and `is_schedule_active` are consistent.
-        if "paused" in update_data:
-            deployment.is_schedule_active = not update_data["paused"]
-        elif "is_schedule_active" in update_data:
-            deployment.paused = not update_data["is_schedule_active"]
+            # Clear schedules to handle their update/creation separately
+            deployment.schedules = None
+        elif not all_provided_have_slugs and all_existing_have_slugs:
+            raise HTTPException(
+                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
+                detail="Please provide a slug for each schedule in your request to ensure schedules are updated correctly.",
+            )
 
         if deployment.work_pool_name:
             # Make sure that deployment is valid before beginning creation process
@@ -263,16 +283,19 @@ async def update_deployment(
         else:
             parameters = existing_deployment.parameters
 
-        # default value for this setting
         enforce_parameter_schema = (
             deployment.enforce_parameter_schema
             if deployment.enforce_parameter_schema is not None
             else existing_deployment.enforce_parameter_schema
         )
-
         if enforce_parameter_schema:
-            # ensure that the new parameters conform to the existing schema
-            if not isinstance(existing_deployment.parameter_openapi_schema, dict):
+            # ensure that the new parameters conform to the proposed schema
+            if deployment.parameter_openapi_schema:
+                openapi_schema = deployment.parameter_openapi_schema
+            else:
+                openapi_schema = existing_deployment.parameter_openapi_schema
+
+            if not isinstance(openapi_schema, dict):
                 raise HTTPException(
                     status.HTTP_409_CONFLICT,
                     detail=(
@@ -284,7 +307,7 @@ async def update_deployment(
             try:
                 validate(
                     parameters,
-                    existing_deployment.parameter_openapi_schema,
+                    openapi_schema,
                     raise_on_error=True,
                     ignore_required=True,
                 )
@@ -299,9 +322,47 @@ async def update_deployment(
                     detail="Invalid schema: Unable to validate schema with circular references.",
                 )
 
+        if deployment.global_concurrency_limit_id:
+            concurrency_limit = (
+                await models.concurrency_limits_v2.read_concurrency_limit(
+                    session=session,
+                    concurrency_limit_id=deployment.global_concurrency_limit_id,
+                )
+            )
+
+            if not concurrency_limit:
+                raise HTTPException(
+                    status_code=status.HTTP_404_NOT_FOUND,
+                    detail="Concurrency limit not found",
+                )
+
         result = await models.deployments.update_deployment(
-            session=session, deployment_id=deployment_id, deployment=deployment
+            session=session,
+            deployment_id=deployment_id,
+            deployment=deployment,
         )
+
+        for schedule in schedules_to_patch:
+            await models.deployments.update_deployment_schedule(
+                session=session,
+                deployment_id=deployment_id,
+                schedule=schedule,
+                deployment_schedule_slug=schedule.slug,
+            )
+        if schedules_to_create:
+            await models.deployments.create_deployment_schedules(
+                session=session,
+                deployment_id=deployment_id,
+                schedules=[
+                    schemas.actions.DeploymentScheduleCreate(
+                        schedule=schedule.schedule,  # type: ignore We will raise above if schedule is not provided
+                        active=schedule.active if schedule.active is not None else True,
+                        slug=schedule.slug,
+                        parameters=schedule.parameters,
+                    )
+                    for schedule in schedules_to_create
+                ],
+            )
     if not result:
         raise HTTPException(status.HTTP_404_NOT_FOUND, detail="Deployment not found.")
 
@@ -323,10 +384,12 @@ async def read_deployment_by_name(
             raise HTTPException(
                 status.HTTP_404_NOT_FOUND, detail="Deployment not found"
             )
-        return schemas.responses.DeploymentResponse.from_orm(deployment)
+        return schemas.responses.DeploymentResponse.model_validate(
+            deployment, from_attributes=True
+        )
 
 
-@router.get("/{id}")
+@router.get("/{id:uuid}")
 async def read_deployment(
     deployment_id: UUID = Path(..., description="The deployment id", alias="id"),
     db: PrefectDBInterface = Depends(provide_database_interface),
@@ -342,19 +405,21 @@ async def read_deployment(
             raise HTTPException(
                 status_code=status.HTTP_404_NOT_FOUND, detail="Deployment not found"
             )
-        return schemas.responses.DeploymentResponse.from_orm(deployment)
+        return schemas.responses.DeploymentResponse.model_validate(
+            deployment, from_attributes=True
+        )
 
 
 @router.post("/filter")
 async def read_deployments(
     limit: int = dependencies.LimitBody(),
     offset: int = Body(0, ge=0),
-    flows: schemas.filters.FlowFilter = None,
-    flow_runs: schemas.filters.FlowRunFilter = None,
-    task_runs: schemas.filters.TaskRunFilter = None,
-    deployments: schemas.filters.DeploymentFilter = None,
-    work_pools: schemas.filters.WorkPoolFilter = None,
-    work_pool_queues: schemas.filters.WorkQueueFilter = None,
+    flows: Optional[schemas.filters.FlowFilter] = None,
+    flow_runs: Optional[schemas.filters.FlowRunFilter] = None,
+    task_runs: Optional[schemas.filters.TaskRunFilter] = None,
+    deployments: Optional[schemas.filters.DeploymentFilter] = None,
+    work_pools: Optional[schemas.filters.WorkPoolFilter] = None,
+    work_pool_queues: Optional[schemas.filters.WorkQueueFilter] = None,
     sort: schemas.sorting.DeploymentSort = Body(
         schemas.sorting.DeploymentSort.NAME_ASC
     ),
@@ -377,23 +442,85 @@ async def read_deployments(
             work_queue_filter=work_pool_queues,
         )
         return [
-            schemas.responses.DeploymentResponse.from_orm(orm_deployment=deployment)
+            schemas.responses.DeploymentResponse.model_validate(
+                deployment, from_attributes=True
+            )
             for deployment in response
         ]
 
 
+@router.post("/paginate")
+async def paginate_deployments(
+    limit: int = dependencies.LimitBody(),
+    page: int = Body(1, ge=1),
+    flows: Optional[schemas.filters.FlowFilter] = None,
+    flow_runs: Optional[schemas.filters.FlowRunFilter] = None,
+    task_runs: Optional[schemas.filters.TaskRunFilter] = None,
+    deployments: Optional[schemas.filters.DeploymentFilter] = None,
+    work_pools: Optional[schemas.filters.WorkPoolFilter] = None,
+    work_pool_queues: Optional[schemas.filters.WorkQueueFilter] = None,
+    sort: schemas.sorting.DeploymentSort = Body(
+        schemas.sorting.DeploymentSort.NAME_ASC
+    ),
+    db: PrefectDBInterface = Depends(provide_database_interface),
+) -> DeploymentPaginationResponse:
+    """
+    Pagination query for flow runs.
+    """
+    offset = (page - 1) * limit
+
+    async with db.session_context() as session:
+        response = await models.deployments.read_deployments(
+            session=session,
+            offset=offset,
+            sort=sort,
+            limit=limit,
+            flow_filter=flows,
+            flow_run_filter=flow_runs,
+            task_run_filter=task_runs,
+            deployment_filter=deployments,
+            work_pool_filter=work_pools,
+            work_queue_filter=work_pool_queues,
+        )
+
+        count = await models.deployments.count_deployments(
+            session=session,
+            flow_filter=flows,
+            flow_run_filter=flow_runs,
+            task_run_filter=task_runs,
+            deployment_filter=deployments,
+            work_pool_filter=work_pools,
+            work_queue_filter=work_pool_queues,
+        )
+
+    results = [
+        schemas.responses.DeploymentResponse.model_validate(
+            deployment, from_attributes=True
+        )
+        for deployment in response
+    ]
+
+    return DeploymentPaginationResponse(
+        results=results,
+        count=count,
+        limit=limit,
+        pages=(count + limit - 1) // limit,
+        page=page,
+    )
+
+
 @router.post("/get_scheduled_flow_runs")
 async def get_scheduled_flow_runs_for_deployments(
-    background_tasks: BackgroundTasks,
-    deployment_ids: List[UUID] = Body(
+    docket: dependencies.Docket,
+    deployment_ids: list[UUID] = Body(
         default=..., description="The deployment IDs to get scheduled runs for"
     ),
-    scheduled_before: DateTimeTZ = Body(
+    scheduled_before: DateTime = Body(
         None, description="The maximum time to look for scheduled flow runs"
     ),
     limit: int = dependencies.LimitBody(),
     db: PrefectDBInterface = Depends(provide_database_interface),
-) -> List[schemas.responses.FlowRunResponse]:
+) -> list[schemas.responses.FlowRunResponse]:
     """
     Get scheduled runs for a set of deployments. Used by a runner to poll for work.
     """
@@ -418,12 +545,13 @@ async def get_scheduled_flow_runs_for_deployments(
         )
 
         flow_run_responses = [
-            schemas.responses.FlowRunResponse.from_orm(orm_flow_run=orm_flow_run)
+            schemas.responses.FlowRunResponse.model_validate(
+                orm_flow_run, from_attributes=True
+            )
             for orm_flow_run in orm_flow_runs
         ]
 
-    background_tasks.add_task(
-        mark_deployments_ready,
+    await docket.add(mark_deployments_ready)(
         deployment_ids=deployment_ids,
     )
 
@@ -432,12 +560,12 @@ async def get_scheduled_flow_runs_for_deployments(
 
 @router.post("/count")
 async def count_deployments(
-    flows: schemas.filters.FlowFilter = None,
-    flow_runs: schemas.filters.FlowRunFilter = None,
-    task_runs: schemas.filters.TaskRunFilter = None,
-    deployments: schemas.filters.DeploymentFilter = None,
-    work_pools: schemas.filters.WorkPoolFilter = None,
-    work_pool_queues: schemas.filters.WorkQueueFilter = None,
+    flows: Optional[schemas.filters.FlowFilter] = None,
+    flow_runs: Optional[schemas.filters.FlowRunFilter] = None,
+    task_runs: Optional[schemas.filters.TaskRunFilter] = None,
+    deployments: Optional[schemas.filters.DeploymentFilter] = None,
+    work_pools: Optional[schemas.filters.WorkPoolFilter] = None,
+    work_pool_queues: Optional[schemas.filters.WorkQueueFilter] = None,
     db: PrefectDBInterface = Depends(provide_database_interface),
 ) -> int:
     """
@@ -455,11 +583,11 @@ async def count_deployments(
         )
 
 
-@router.delete("/{id}", status_code=status.HTTP_204_NO_CONTENT)
+@router.delete("/{id:uuid}", status_code=status.HTTP_204_NO_CONTENT)
 async def delete_deployment(
     deployment_id: UUID = Path(..., description="The deployment id", alias="id"),
     db: PrefectDBInterface = Depends(provide_database_interface),
-):
+) -> None:
     """
     Delete a deployment by id.
     """
@@ -473,16 +601,21 @@ async def delete_deployment(
         )
 
 
-@router.post("/{id}/schedule")
+@router.post("/{id:uuid}/schedule")
 async def schedule_deployment(
     deployment_id: UUID = Path(..., description="The deployment id", alias="id"),
-    start_time: DateTimeTZ = Body(None, description="The earliest date to schedule"),
-    end_time: DateTimeTZ = Body(None, description="The latest date to schedule"),
-    min_time: datetime.timedelta = Body(
+    start_time: datetime.datetime = Body(
+        None, description="The earliest date to schedule"
+    ),
+    end_time: datetime.datetime = Body(None, description="The latest date to schedule"),
+    # Workaround for the fact that FastAPI does not let us configure ser_json_timedelta
+    # to represent timedeltas as floats in JSON.
+    min_time: float = Body(
         None,
         description=(
             "Runs will be scheduled until at least this long after the `start_time`"
         ),
+        json_schema_extra={"format": "time-delta"},
     ),
     min_runs: int = Body(None, description="The minimum number of runs to schedule"),
     max_runs: int = Body(None, description="The maximum number of runs to schedule"),
@@ -501,6 +634,9 @@ async def schedule_deployment(
         - At least `min_runs` runs will be generated
         - Runs will be generated until at least `start_time + min_time` is reached
     """
+    if isinstance(min_time, float):
+        min_time = datetime.timedelta(seconds=min_time)
+
     async with db.session_context(begin_transaction=True) as session:
         await models.deployments.schedule_runs(
             session=session,
@@ -513,7 +649,6 @@ async def schedule_deployment(
         )
 
 
-@router.post("/{id:uuid}/set_schedule_active")  # Legacy route
 @router.post("/{id:uuid}/resume_deployment")
 async def resume_deployment(
     deployment_id: UUID = Path(..., description="The deployment id", alias="id"),
@@ -530,22 +665,9 @@ async def resume_deployment(
             raise HTTPException(
                 status_code=status.HTTP_404_NOT_FOUND, detail="Deployment not found"
             )
-        deployment.is_schedule_active = True
         deployment.paused = False
 
-        # Ensure that we're updating the replicated schedule's `active` field,
-        # if there is only a single schedule. This is support for legacy
-        # clients.
 
-        number_of_schedules = len(deployment.schedules)
-
-        if number_of_schedules == 1:
-            deployment.schedules[0].active = True
-        elif number_of_schedules > 1:
-            raise _multiple_schedules_error(deployment_id)
-
-
-@router.post("/{id:uuid}/set_schedule_inactive")  # Legacy route
 @router.post("/{id:uuid}/pause_deployment")
 async def pause_deployment(
     deployment_id: UUID = Path(..., description="The deployment id", alias="id"),
@@ -563,20 +685,8 @@ async def pause_deployment(
             raise HTTPException(
                 status_code=status.HTTP_404_NOT_FOUND, detail="Deployment not found"
             )
-        deployment.is_schedule_active = False
         deployment.paused = True
 
-        # Ensure that we're updating the replicated schedule's `active` field,
-        # if there is only a single schedule. This is support for legacy
-        # clients.
-
-        number_of_schedules = len(deployment.schedules)
-
-        if number_of_schedules == 1:
-            deployment.schedules[0].active = False
-        elif number_of_schedules > 1:
-            raise _multiple_schedules_error(deployment_id)
-
         # commit here to make the inactive schedule "visible" to the scheduler service
         await session.commit()
 
@@ -590,7 +700,7 @@ async def pause_deployment(
         await session.commit()
 
 
-@router.post("/{id}/create_flow_run")
+@router.post("/{id:uuid}/create_flow_run")
 async def create_flow_run_from_deployment(
     flow_run: schemas.actions.DeploymentFlowRunCreate,
     deployment_id: UUID = Path(..., description="The deployment id", alias="id"),
@@ -675,6 +785,12 @@ async def create_flow_run_from_deployment(
 
         if flow_run.work_queue_name:
             # can't mutate the ORM model or else it will commit the changes back
+            if deployment.work_queue is None or deployment.work_queue.work_pool is None:
+                raise HTTPException(
+                    status_code=status.HTTP_400_BAD_REQUEST,
+                    detail=f"Cannot create flow run in work queue {flow_run.work_queue_name} because deployment {deployment_id} is not associated with a work pool. Please remove work_pool_name and try again.",
+                )
+
             work_queue_id = await worker_lookups._get_work_queue_id_from_name(
                 session=session,
                 work_pool_name=deployment.work_queue.work_pool.name,
@@ -685,7 +801,7 @@ async def create_flow_run_from_deployment(
 
         # hydrate the input model into a full flow run / state model
         flow_run = schemas.core.FlowRun(
-            **flow_run.dict(
+            **flow_run.model_dump(
                 exclude={
                     "parameters",
                     "tags",
@@ -711,17 +827,19 @@ async def create_flow_run_from_deployment(
         if not flow_run.state:
             flow_run.state = schemas.states.Scheduled()
 
-        now = pendulum.now("UTC")
+        right_now = now("UTC")
         model = await models.flow_runs.create_flow_run(
             session=session, flow_run=flow_run
         )
-        if model.created >= now:
+        if model.created >= right_now:
             response.status_code = status.HTTP_201_CREATED
-        return schemas.responses.FlowRunResponse.from_orm(model)
+        return schemas.responses.FlowRunResponse.model_validate(
+            model, from_attributes=True
+        )
 
 
 # DEPRECATED
-@router.get("/{id}/work_queue_check", deprecated=True)
+@router.get("/{id:uuid}/work_queue_check", deprecated=True)
 async def work_queue_check_for_deployment(
     deployment_id: UUID = Path(..., description="The deployment id", alias="id"),
     db: PrefectDBInterface = Depends(provide_database_interface),
@@ -747,7 +865,7 @@ async def work_queue_check_for_deployment(
     return work_queues
 
 
-@router.get("/{id}/schedules")
+@router.get("/{id:uuid}/schedules")
 async def read_deployment_schedules(
     deployment_id: UUID = Path(..., description="The deployment id", alias="id"),
     db: PrefectDBInterface = Depends(provide_database_interface),
@@ -768,7 +886,7 @@ async def read_deployment_schedules(
         )
 
 
-@router.post("/{id}/schedules", status_code=status.HTTP_201_CREATED)
+@router.post("/{id:uuid}/schedules", status_code=status.HTTP_201_CREATED)
 async def create_deployment_schedules(
     deployment_id: UUID = Path(..., description="The deployment id", alias="id"),
     schedules: List[schemas.actions.DeploymentScheduleCreate] = Body(
@@ -786,16 +904,25 @@ async def create_deployment_schedules(
                 status.HTTP_404_NOT_FOUND, detail="Deployment not found."
             )
 
-        created = await models.deployments.create_deployment_schedules(
-            session=session,
-            deployment_id=deployment.id,
-            schedules=schedules,
-        )
-
+        try:
+            created = await models.deployments.create_deployment_schedules(
+                session=session,
+                deployment_id=deployment.id,
+                schedules=schedules,
+            )
+        except sa.exc.IntegrityError as e:
+            if "duplicate key value violates unique constraint" in str(e):
+                raise HTTPException(
+                    status.HTTP_409_CONFLICT,
+                    detail="Schedule slugs must be unique within a deployment.",
+                )
+            raise
         return created
 
 
-@router.patch("/{id}/schedules/{schedule_id}", status_code=status.HTTP_204_NO_CONTENT)
+@router.patch(
+    "/{id:uuid}/schedules/{schedule_id:uuid}", status_code=status.HTTP_204_NO_CONTENT
+)
 async def update_deployment_schedule(
     deployment_id: UUID = Path(..., description="The deployment id", alias="id"),
     schedule_id: UUID = Path(..., description="The schedule id", alias="schedule_id"),
@@ -803,7 +930,7 @@ async def update_deployment_schedule(
         default=..., description="The updated schedule"
     ),
     db: PrefectDBInterface = Depends(provide_database_interface),
-):
+) -> None:
     async with db.session_context(begin_transaction=True) as session:
         deployment = await models.deployments.read_deployment(
             session=session, deployment_id=deployment_id
@@ -828,15 +955,18 @@ async def update_deployment_schedule(
             session=session,
             deployment_id=deployment_id,
             auto_scheduled_only=True,
+            future_only=True,
         )
 
 
-@router.delete("/{id}/schedules/{schedule_id}", status_code=status.HTTP_204_NO_CONTENT)
+@router.delete(
+    "/{id:uuid}/schedules/{schedule_id:uuid}", status_code=status.HTTP_204_NO_CONTENT
+)
 async def delete_deployment_schedule(
     deployment_id: UUID = Path(..., description="The deployment id", alias="id"),
     schedule_id: UUID = Path(..., description="The schedule id", alias="schedule_id"),
     db: PrefectDBInterface = Depends(provide_database_interface),
-):
+) -> None:
     async with db.session_context(begin_transaction=True) as session:
         deployment = await models.deployments.read_deployment(
             session=session, deployment_id=deployment_id
