@@ -2,97 +2,137 @@
 Routes for interacting with flow run objects.
 """
 
+import csv
 import datetime
-from typing import Any, Dict, List, Optional
+import io
+from typing import TYPE_CHECKING, Any, Dict, List, Optional
 from uuid import UUID
 
 import orjson
-import pendulum
 import sqlalchemy as sa
-from prefect._vendor.fastapi import (
+from docket import Depends as DocketDepends
+from docket import Retry
+from fastapi import (
     Body,
     Depends,
     HTTPException,
     Path,
     Query,
     Response,
-    status,
 )
-from prefect._vendor.fastapi.responses import ORJSONResponse, PlainTextResponse
+from fastapi.encoders import jsonable_encoder
+from fastapi.responses import ORJSONResponse, PlainTextResponse, StreamingResponse
 from sqlalchemy.exc import IntegrityError
 
 import prefect.server.api.dependencies as dependencies
 import prefect.server.models as models
 import prefect.server.schemas as schemas
+from prefect._internal.compatibility.starlette import status
 from prefect.logging import get_logger
 from prefect.server.api.run_history import run_history
 from prefect.server.api.validation import validate_job_variables_for_deployment_flow_run
-from prefect.server.database.dependencies import provide_database_interface
-from prefect.server.database.interface import PrefectDBInterface
+from prefect.server.api.workers import WorkerLookups
+from prefect.server.database import PrefectDBInterface, provide_database_interface
 from prefect.server.exceptions import FlowRunGraphTooLarge
 from prefect.server.models.flow_runs import (
     DependencyResult,
     read_flow_run_graph,
 )
 from prefect.server.orchestration import dependencies as orchestration_dependencies
-from prefect.server.orchestration.policies import BaseOrchestrationPolicy
+from prefect.server.orchestration.policies import (
+    FlowRunOrchestrationPolicy,
+    TaskRunOrchestrationPolicy,
+)
 from prefect.server.schemas.graph import Graph
-from prefect.server.schemas.responses import OrchestrationResult
-from prefect.server.utilities.schemas import DateTimeTZ
+from prefect.server.schemas.responses import (
+    FlowRunPaginationResponse,
+    OrchestrationResult,
+)
 from prefect.server.utilities.server import PrefectRouter
+from prefect.types import DateTime
+from prefect.types._datetime import earliest_possible_datetime, now
 from prefect.utilities import schema_tools
 
-logger = get_logger("server.api")
+if TYPE_CHECKING:
+    import logging
 
-router = PrefectRouter(prefix="/flow_runs", tags=["Flow Runs"])
+logger: "logging.Logger" = get_logger("server.api")
+
+router: PrefectRouter = PrefectRouter(prefix="/flow_runs", tags=["Flow Runs"])
 
 
 @router.post("/")
 async def create_flow_run(
     flow_run: schemas.actions.FlowRunCreate,
     db: PrefectDBInterface = Depends(provide_database_interface),
-    response: Response = None,
+    response: Response = None,  # type: ignore
     created_by: Optional[schemas.core.CreatedBy] = Depends(dependencies.get_created_by),
     orchestration_parameters: Dict[str, Any] = Depends(
         orchestration_dependencies.provide_flow_orchestration_parameters
     ),
-    api_version=Depends(dependencies.provide_request_api_version),
+    api_version: str = Depends(dependencies.provide_request_api_version),
+    worker_lookups: WorkerLookups = Depends(WorkerLookups),
 ) -> schemas.responses.FlowRunResponse:
     """
     Create a flow run. If a flow run with the same flow_id and
     idempotency key already exists, the existing flow run will be returned.
 
     If no state is provided, the flow run will be created in a PENDING state.
+
+    For more information, see https://docs.prefect.io/v3/concepts/flows.
     """
     # hydrate the input model into a full flow run / state model
-    flow_run = schemas.core.FlowRun(**flow_run.dict(), created_by=created_by)
+    flow_run_object = schemas.core.FlowRun(
+        **flow_run.model_dump(), created_by=created_by
+    )
 
     # pass the request version to the orchestration engine to support compatibility code
     orchestration_parameters.update({"api-version": api_version})
 
-    if not flow_run.state:
-        flow_run.state = schemas.states.Pending()
+    if not flow_run_object.state:
+        flow_run_object.state = schemas.states.Pending()
 
-    now = pendulum.now("UTC")
+    right_now = now("UTC")
 
     async with db.session_context(begin_transaction=True) as session:
+        if flow_run.work_pool_name:
+            if flow_run.work_queue_name:
+                work_queue_id = await worker_lookups._get_work_queue_id_from_name(
+                    session=session,
+                    work_pool_name=flow_run.work_pool_name,
+                    work_queue_name=flow_run.work_queue_name,
+                )
+            else:
+                work_queue_id = (
+                    await worker_lookups._get_default_work_queue_id_from_work_pool_name(
+                        session=session,
+                        work_pool_name=flow_run.work_pool_name,
+                    )
+                )
+        else:
+            work_queue_id = None
+
+        flow_run_object.work_queue_id = work_queue_id
+
         model = await models.flow_runs.create_flow_run(
             session=session,
-            flow_run=flow_run,
+            flow_run=flow_run_object,
             orchestration_parameters=orchestration_parameters,
         )
-        if model.created >= now:
+        if model.created >= right_now:
             response.status_code = status.HTTP_201_CREATED
 
-        return schemas.responses.FlowRunResponse.from_orm(model)
+        return schemas.responses.FlowRunResponse.model_validate(
+            model, from_attributes=True
+        )
 
 
-@router.patch("/{id}", status_code=status.HTTP_204_NO_CONTENT)
+@router.patch("/{id:uuid}", status_code=status.HTTP_204_NO_CONTENT)
 async def update_flow_run(
     flow_run: schemas.actions.FlowRunUpdate,
     flow_run_id: UUID = Path(..., description="The flow run id", alias="id"),
     db: PrefectDBInterface = Depends(provide_database_interface),
-):
+) -> None:
     """
     Updates a flow run.
     """
@@ -143,12 +183,12 @@ async def update_flow_run(
 
 @router.post("/count")
 async def count_flow_runs(
-    flows: schemas.filters.FlowFilter = None,
-    flow_runs: schemas.filters.FlowRunFilter = None,
-    task_runs: schemas.filters.TaskRunFilter = None,
-    deployments: schemas.filters.DeploymentFilter = None,
-    work_pools: schemas.filters.WorkPoolFilter = None,
-    work_pool_queues: schemas.filters.WorkQueueFilter = None,
+    flows: Optional[schemas.filters.FlowFilter] = None,
+    flow_runs: Optional[schemas.filters.FlowRunFilter] = None,
+    task_runs: Optional[schemas.filters.TaskRunFilter] = None,
+    deployments: Optional[schemas.filters.DeploymentFilter] = None,
+    work_pools: Optional[schemas.filters.WorkPoolFilter] = None,
+    work_pool_queues: Optional[schemas.filters.WorkQueueFilter] = None,
     db: PrefectDBInterface = Depends(provide_database_interface),
 ) -> int:
     """
@@ -206,6 +246,7 @@ async def average_flow_run_lateness(
             base_query = db.FlowRun.estimated_start_time_delta
 
         query = await models.flow_runs._apply_flow_run_filters(
+            db,
             sa.select(sa.func.avg(base_query)),
             flow_filter=flows,
             flow_run_filter=flow_runs,
@@ -228,26 +269,33 @@ async def average_flow_run_lateness(
 
 @router.post("/history")
 async def flow_run_history(
-    history_start: DateTimeTZ = Body(..., description="The history's start time."),
-    history_end: DateTimeTZ = Body(..., description="The history's end time."),
-    history_interval: datetime.timedelta = Body(
+    history_start: DateTime = Body(..., description="The history's start time."),
+    history_end: DateTime = Body(..., description="The history's end time."),
+    # Workaround for the fact that FastAPI does not let us configure ser_json_timedelta
+    # to represent timedeltas as floats in JSON.
+    history_interval: float = Body(
         ...,
         description=(
             "The size of each history interval, in seconds. Must be at least 1 second."
         ),
+        json_schema_extra={"format": "time-delta"},
         alias="history_interval_seconds",
     ),
-    flows: schemas.filters.FlowFilter = None,
-    flow_runs: schemas.filters.FlowRunFilter = None,
-    task_runs: schemas.filters.TaskRunFilter = None,
-    deployments: schemas.filters.DeploymentFilter = None,
-    work_pools: schemas.filters.WorkPoolFilter = None,
-    work_queues: schemas.filters.WorkQueueFilter = None,
+    flows: Optional[schemas.filters.FlowFilter] = None,
+    flow_runs: Optional[schemas.filters.FlowRunFilter] = None,
+    task_runs: Optional[schemas.filters.TaskRunFilter] = None,
+    deployments: Optional[schemas.filters.DeploymentFilter] = None,
+    work_pools: Optional[schemas.filters.WorkPoolFilter] = None,
+    work_queues: Optional[schemas.filters.WorkQueueFilter] = None,
     db: PrefectDBInterface = Depends(provide_database_interface),
 ) -> List[schemas.responses.HistoryResponse]:
     """
     Query for flow run history data across a given range and interval.
     """
+    if isinstance(history_interval, float):
+        history_interval = datetime.timedelta(seconds=history_interval)
+
+    assert isinstance(history_interval, datetime.timedelta)
     if history_interval < datetime.timedelta(seconds=1):
         raise HTTPException(
             status.HTTP_422_UNPROCESSABLE_ENTITY,
@@ -270,7 +318,7 @@ async def flow_run_history(
         )
 
 
-@router.get("/{id}")
+@router.get("/{id:uuid}")
 async def read_flow_run(
     flow_run_id: UUID = Path(..., description="The flow run id", alias="id"),
     db: PrefectDBInterface = Depends(provide_database_interface),
@@ -284,10 +332,12 @@ async def read_flow_run(
         )
         if not flow_run:
             raise HTTPException(status.HTTP_404_NOT_FOUND, detail="Flow run not found")
-        return schemas.responses.FlowRunResponse.from_orm(flow_run)
+        return schemas.responses.FlowRunResponse.model_validate(
+            flow_run, from_attributes=True
+        )
 
 
-@router.get("/{id}/graph")
+@router.get("/{id:uuid}/graph", tags=["Flow Run Graph"])
 async def read_flow_run_graph_v1(
     flow_run_id: UUID = Path(..., description="The flow run id", alias="id"),
     db: PrefectDBInterface = Depends(provide_database_interface),
@@ -301,11 +351,11 @@ async def read_flow_run_graph_v1(
         )
 
 
-@router.get("/{id:uuid}/graph-v2")
+@router.get("/{id:uuid}/graph-v2", tags=["Flow Run Graph"])
 async def read_flow_run_graph_v2(
     flow_run_id: UUID = Path(..., description="The flow run id", alias="id"),
     since: datetime.datetime = Query(
-        datetime.datetime.min,
+        default=jsonable_encoder(earliest_possible_datetime()),
         description="Only include runs that start or end after this time.",
     ),
     db: PrefectDBInterface = Depends(provide_database_interface),
@@ -327,27 +377,28 @@ async def read_flow_run_graph_v2(
             )
 
 
-@router.post("/{id}/resume")
+@router.post("/{id:uuid}/resume")
 async def resume_flow_run(
+    response: Response,
     flow_run_id: UUID = Path(..., description="The flow run id", alias="id"),
     db: PrefectDBInterface = Depends(provide_database_interface),
-    run_input: Optional[Dict] = Body(default=None, embed=True),
-    response: Response = None,
-    flow_policy: BaseOrchestrationPolicy = Depends(
+    run_input: Optional[dict[str, Any]] = Body(default=None, embed=True),
+    flow_policy: type[FlowRunOrchestrationPolicy] = Depends(
         orchestration_dependencies.provide_flow_policy
     ),
-    task_policy: BaseOrchestrationPolicy = Depends(
+    task_policy: type[TaskRunOrchestrationPolicy] = Depends(
         orchestration_dependencies.provide_task_policy
     ),
     orchestration_parameters: Dict[str, Any] = Depends(
         orchestration_dependencies.provide_flow_orchestration_parameters
     ),
-    api_version=Depends(dependencies.provide_request_api_version),
+    api_version: str = Depends(dependencies.provide_request_api_version),
+    client_version: Optional[str] = Depends(dependencies.get_prefect_client_version),
 ) -> OrchestrationResult:
     """
     Resume a paused flow run.
     """
-    now = pendulum.now("UTC")
+    right_now = now("UTC")
 
     async with db.session_context(begin_transaction=True) as session:
         flow_run = await models.flow_runs.read_flow_run(session, flow_run_id)
@@ -435,10 +486,11 @@ async def resume_flow_run(
                 session=session,
                 flow_run_id=flow_run_id,
                 state=schemas.states.Scheduled(
-                    name="Resuming", scheduled_time=pendulum.now("UTC")
+                    name="Resuming", scheduled_time=now("UTC")
                 ),
                 flow_policy=flow_policy,
                 orchestration_parameters=orchestration_parameters,
+                client_version=client_version,
             )
         else:
             orchestration_result = await models.flow_runs.set_flow_run_state(
@@ -447,6 +499,7 @@ async def resume_flow_run(
                 state=schemas.states.Running(),
                 flow_policy=flow_policy,
                 orchestration_parameters=orchestration_parameters,
+                client_version=client_version,
             )
 
         if (
@@ -466,7 +519,10 @@ async def resume_flow_run(
             )
 
         # set the 201 if a new state was created
-        if orchestration_result.state and orchestration_result.state.timestamp >= now:
+        if (
+            orchestration_result.state
+            and orchestration_result.state.timestamp >= right_now
+        ):
             response.status_code = status.HTTP_201_CREATED
         else:
             response.status_code = status.HTTP_200_OK
@@ -479,12 +535,12 @@ async def read_flow_runs(
     sort: schemas.sorting.FlowRunSort = Body(schemas.sorting.FlowRunSort.ID_DESC),
     limit: int = dependencies.LimitBody(),
     offset: int = Body(0, ge=0),
-    flows: schemas.filters.FlowFilter = None,
-    flow_runs: schemas.filters.FlowRunFilter = None,
-    task_runs: schemas.filters.TaskRunFilter = None,
-    deployments: schemas.filters.DeploymentFilter = None,
-    work_pools: schemas.filters.WorkPoolFilter = None,
-    work_pool_queues: schemas.filters.WorkQueueFilter = None,
+    flows: Optional[schemas.filters.FlowFilter] = None,
+    flow_runs: Optional[schemas.filters.FlowRunFilter] = None,
+    task_runs: Optional[schemas.filters.TaskRunFilter] = None,
+    deployments: Optional[schemas.filters.DeploymentFilter] = None,
+    work_pools: Optional[schemas.filters.WorkPoolFilter] = None,
+    work_pool_queues: Optional[schemas.filters.WorkQueueFilter] = None,
     db: PrefectDBInterface = Depends(provide_database_interface),
 ) -> List[schemas.responses.FlowRunResponse]:
     """
@@ -509,17 +565,20 @@ async def read_flow_runs(
         # In particular, the FastAPI encoder is very slow for large, nested objects.
         # See: https://github.com/tiangolo/fastapi/issues/1224
         encoded = [
-            schemas.responses.FlowRunResponse.from_orm(fr).dict(json_compatible=True)
+            schemas.responses.FlowRunResponse.model_validate(
+                fr, from_attributes=True
+            ).model_dump(mode="json")
             for fr in db_flow_runs
         ]
         return ORJSONResponse(content=encoded)
 
 
-@router.delete("/{id}", status_code=status.HTTP_204_NO_CONTENT)
+@router.delete("/{id:uuid}", status_code=status.HTTP_204_NO_CONTENT)
 async def delete_flow_run(
+    docket: dependencies.Docket,
     flow_run_id: UUID = Path(..., description="The flow run id", alias="id"),
     db: PrefectDBInterface = Depends(provide_database_interface),
-):
+) -> None:
     """
     Delete a flow run by id.
     """
@@ -531,10 +590,27 @@ async def delete_flow_run(
         raise HTTPException(
             status_code=status.HTTP_404_NOT_FOUND, detail="Flow run not found"
         )
+    await docket.add(delete_flow_run_logs)(flow_run_id=flow_run_id)
 
 
-@router.post("/{id}/set_state")
+async def delete_flow_run_logs(
+    *,
+    db: PrefectDBInterface = DocketDepends(provide_database_interface),
+    flow_run_id: UUID,
+    retry: Retry = Retry(attempts=5, delay=datetime.timedelta(seconds=0.5)),
+) -> None:
+    async with db.session_context(begin_transaction=True) as session:
+        await models.logs.delete_logs(
+            session=session,
+            log_filter=schemas.filters.LogFilter(
+                flow_run_id=schemas.filters.LogFilterFlowRunId(any_=[flow_run_id])
+            ),
+        )
+
+
+@router.post("/{id:uuid}/set_state")
 async def set_flow_run_state(
+    response: Response,
     flow_run_id: UUID = Path(..., description="The flow run id", alias="id"),
     state: schemas.actions.StateCreate = Body(..., description="The intended state."),
     force: bool = Body(
@@ -545,21 +621,21 @@ async def set_flow_run_state(
         ),
     ),
     db: PrefectDBInterface = Depends(provide_database_interface),
-    response: Response = None,
-    flow_policy: BaseOrchestrationPolicy = Depends(
+    flow_policy: type[FlowRunOrchestrationPolicy] = Depends(
         orchestration_dependencies.provide_flow_policy
     ),
     orchestration_parameters: Dict[str, Any] = Depends(
         orchestration_dependencies.provide_flow_orchestration_parameters
     ),
-    api_version=Depends(dependencies.provide_request_api_version),
+    api_version: str = Depends(dependencies.provide_request_api_version),
+    client_version: Optional[str] = Depends(dependencies.get_prefect_client_version),
 ) -> OrchestrationResult:
     """Set a flow run state, invoking any orchestration rules."""
 
     # pass the request version to the orchestration engine to support compatibility code
     orchestration_parameters.update({"api-version": api_version})
 
-    now = pendulum.now("UTC")
+    right_now = now("UTC")
 
     # create the state
     async with db.session_context(
@@ -569,14 +645,15 @@ async def set_flow_run_state(
             session=session,
             flow_run_id=flow_run_id,
             # convert to a full State object
-            state=schemas.states.State.parse_obj(state),
+            state=schemas.states.State.model_validate(state),
             force=force,
             flow_policy=flow_policy,
             orchestration_parameters=orchestration_parameters,
+            client_version=client_version,
         )
 
     # set the 201 if a new state was created
-    if orchestration_result.state and orchestration_result.state.timestamp >= now:
+    if orchestration_result.state and orchestration_result.state.timestamp >= right_now:
         response.status_code = status.HTTP_201_CREATED
     else:
         response.status_code = status.HTTP_200_OK
@@ -584,14 +661,14 @@ async def set_flow_run_state(
     return orchestration_result
 
 
-@router.post("/{id}/input", status_code=status.HTTP_201_CREATED)
+@router.post("/{id:uuid}/input", status_code=status.HTTP_201_CREATED)
 async def create_flow_run_input(
     flow_run_id: UUID = Path(..., description="The flow run id", alias="id"),
     key: str = Body(..., description="The input key"),
     value: bytes = Body(..., description="The value of the input"),
     sender: Optional[str] = Body(None, description="The sender of the input"),
     db: PrefectDBInterface = Depends(provide_database_interface),
-):
+) -> None:
     """
     Create a key/value input for a flow run.
     """
@@ -620,7 +697,7 @@ async def create_flow_run_input(
                 )
 
 
-@router.post("/{id}/input/filter")
+@router.post("/{id:uuid}/input/filter")
 async def filter_flow_run_input(
     flow_run_id: UUID = Path(..., description="The flow run id", alias="id"),
     prefix: str = Body(..., description="The input key prefix", embed=True),
@@ -645,7 +722,7 @@ async def filter_flow_run_input(
         )
 
 
-@router.get("/{id}/input/{key}")
+@router.get("/{id:uuid}/input/{key}")
 async def read_flow_run_input(
     flow_run_id: UUID = Path(..., description="The flow run id", alias="id"),
     key: str = Path(..., description="The input key", alias="key"),
@@ -668,12 +745,12 @@ async def read_flow_run_input(
         )
 
 
-@router.delete("/{id}/input/{key}", status_code=status.HTTP_204_NO_CONTENT)
+@router.delete("/{id:uuid}/input/{key}", status_code=status.HTTP_204_NO_CONTENT)
 async def delete_flow_run_input(
     flow_run_id: UUID = Path(..., description="The flow run id", alias="id"),
     key: str = Path(..., description="The input key", alias="key"),
     db: PrefectDBInterface = Depends(provide_database_interface),
-):
+) -> None:
     """
     Delete a flow run input
     """
@@ -688,3 +765,151 @@ async def delete_flow_run_input(
             raise HTTPException(
                 status_code=status.HTTP_404_NOT_FOUND, detail="Flow run input not found"
             )
+
+
+@router.post("/paginate", response_class=ORJSONResponse)
+async def paginate_flow_runs(
+    sort: schemas.sorting.FlowRunSort = Body(schemas.sorting.FlowRunSort.ID_DESC),
+    limit: int = dependencies.LimitBody(),
+    page: int = Body(1, ge=1),
+    flows: Optional[schemas.filters.FlowFilter] = None,
+    flow_runs: Optional[schemas.filters.FlowRunFilter] = None,
+    task_runs: Optional[schemas.filters.TaskRunFilter] = None,
+    deployments: Optional[schemas.filters.DeploymentFilter] = None,
+    work_pools: Optional[schemas.filters.WorkPoolFilter] = None,
+    work_pool_queues: Optional[schemas.filters.WorkQueueFilter] = None,
+    db: PrefectDBInterface = Depends(provide_database_interface),
+) -> FlowRunPaginationResponse:
+    """
+    Pagination query for flow runs.
+    """
+    offset = (page - 1) * limit
+
+    async with db.session_context() as session:
+        runs = await models.flow_runs.read_flow_runs(
+            session=session,
+            flow_filter=flows,
+            flow_run_filter=flow_runs,
+            task_run_filter=task_runs,
+            deployment_filter=deployments,
+            work_pool_filter=work_pools,
+            work_queue_filter=work_pool_queues,
+            offset=offset,
+            limit=limit,
+            sort=sort,
+        )
+
+        count = await models.flow_runs.count_flow_runs(
+            session=session,
+            flow_filter=flows,
+            flow_run_filter=flow_runs,
+            task_run_filter=task_runs,
+            deployment_filter=deployments,
+            work_pool_filter=work_pools,
+            work_queue_filter=work_pool_queues,
+        )
+
+        # Instead of relying on fastapi.encoders.jsonable_encoder to convert the
+        # response to JSON, we do so more efficiently ourselves.
+        # In particular, the FastAPI encoder is very slow for large, nested objects.
+        # See: https://github.com/tiangolo/fastapi/issues/1224
+        results = [
+            schemas.responses.FlowRunResponse.model_validate(
+                run, from_attributes=True
+            ).model_dump(mode="json")
+            for run in runs
+        ]
+
+        response = FlowRunPaginationResponse(
+            results=results,
+            count=count,
+            limit=limit,
+            pages=(count + limit - 1) // limit,
+            page=page,
+        ).model_dump(mode="json")
+
+        return ORJSONResponse(content=response)
+
+
+FLOW_RUN_LOGS_DOWNLOAD_PAGE_LIMIT = 1000
+
+
+@router.get("/{id:uuid}/logs/download")
+async def download_logs(
+    flow_run_id: UUID = Path(..., description="The flow run id", alias="id"),
+    db: PrefectDBInterface = Depends(provide_database_interface),
+) -> StreamingResponse:
+    """
+    Download all flow run logs as a CSV file, collecting all logs until there are no more logs to retrieve.
+    """
+    async with db.session_context() as session:
+        flow_run = await models.flow_runs.read_flow_run(
+            session=session, flow_run_id=flow_run_id
+        )
+
+        if not flow_run:
+            raise HTTPException(status.HTTP_404_NOT_FOUND, detail="Flow run not found")
+
+        async def generate():
+            data = io.StringIO()
+            csv_writer = csv.writer(data)
+            csv_writer.writerow(
+                ["timestamp", "level", "flow_run_id", "task_run_id", "message"]
+            )
+
+            offset = 0
+            limit = FLOW_RUN_LOGS_DOWNLOAD_PAGE_LIMIT
+
+            while True:
+                results = await models.logs.read_logs(
+                    session=session,
+                    log_filter=schemas.filters.LogFilter(
+                        flow_run_id={"any_": [flow_run_id]}
+                    ),
+                    offset=offset,
+                    limit=limit,
+                    sort=schemas.sorting.LogSort.TIMESTAMP_ASC,
+                )
+
+                if not results:
+                    break
+
+                offset += limit
+
+                for log in results:
+                    csv_writer.writerow(
+                        [
+                            log.timestamp,
+                            log.level,
+                            log.flow_run_id,
+                            log.task_run_id,
+                            log.message,
+                        ]
+                    )
+                    data.seek(0)
+                    yield data.read()
+                    data.seek(0)
+                    data.truncate(0)
+
+        return StreamingResponse(
+            generate(),
+            media_type="text/csv",
+            headers={
+                "Content-Disposition": f"attachment; filename={flow_run.name}-logs.csv"
+            },
+        )
+
+
+@router.patch("/{id:uuid}/labels", status_code=status.HTTP_204_NO_CONTENT)
+async def update_flow_run_labels(
+    flow_run_id: UUID = Path(..., description="The flow run id", alias="id"),
+    labels: Dict[str, Any] = Body(..., description="The labels to update"),
+    db: PrefectDBInterface = Depends(provide_database_interface),
+) -> None:
+    """
+    Update the labels of a flow run.
+    """
+    async with db.session_context(begin_transaction=True) as session:
+        await models.flow_runs.update_flow_run_labels(
+            session=session, flow_run_id=flow_run_id, labels=labels
+        )
