@@ -5,9 +5,7 @@ Routes for interacting with work queue objects.
 from typing import List, Optional
 from uuid import UUID
 
-import sqlalchemy as sa
-from prefect._vendor.fastapi import (
-    BackgroundTasks,
+from fastapi import (
     Body,
     Depends,
     Header,
@@ -15,22 +13,25 @@ from prefect._vendor.fastapi import (
     Path,
     status,
 )
+from sqlalchemy.exc import IntegrityError
 
 import prefect.server.api.dependencies as dependencies
 import prefect.server.models as models
 import prefect.server.schemas as schemas
-from prefect.server.database.dependencies import db_injector, provide_database_interface
-from prefect.server.database.interface import PrefectDBInterface
+from prefect.server.database import (
+    PrefectDBInterface,
+    provide_database_interface,
+)
 from prefect.server.models.deployments import mark_deployments_ready
 from prefect.server.models.work_queues import (
     emit_work_queue_status_event,
     mark_work_queues_ready,
 )
 from prefect.server.schemas.statuses import WorkQueueStatus
-from prefect.server.utilities.schemas import DateTimeTZ
 from prefect.server.utilities.server import PrefectRouter
+from prefect.types import DateTime
 
-router = PrefectRouter(prefix="/work_queues", tags=["Work Queues"])
+router: PrefectRouter = PrefectRouter(prefix="/work_queues", tags=["Work Queues"])
 
 
 @router.post("/", status_code=status.HTTP_201_CREATED)
@@ -43,6 +44,8 @@ async def create_work_queue(
 
     If a work queue with the same name already exists, an error
     will be raised.
+
+    For more information, see https://docs.prefect.io/v3/concepts/work-pools#work-queues.
     """
 
     try:
@@ -50,21 +53,23 @@ async def create_work_queue(
             model = await models.work_queues.create_work_queue(
                 session=session, work_queue=work_queue
             )
-    except sa.exc.IntegrityError:
+    except IntegrityError:
         raise HTTPException(
             status_code=status.HTTP_409_CONFLICT,
             detail="A work queue with this name already exists.",
         )
 
-    return schemas.responses.WorkQueueResponse.from_orm(model)
+    return schemas.responses.WorkQueueResponse.model_validate(
+        model, from_attributes=True
+    )
 
 
-@router.patch("/{id}", status_code=status.HTTP_204_NO_CONTENT)
+@router.patch("/{id:uuid}", status_code=status.HTTP_204_NO_CONTENT)
 async def update_work_queue(
     work_queue: schemas.actions.WorkQueueUpdate,
     work_queue_id: UUID = Path(..., description="The work queue id", alias="id"),
     db: PrefectDBInterface = Depends(provide_database_interface),
-):
+) -> None:
     """
     Updates an existing work queue.
     """
@@ -97,10 +102,12 @@ async def read_work_queue_by_name(
         raise HTTPException(
             status_code=status.HTTP_404_NOT_FOUND, detail="work queue not found"
         )
-    return schemas.responses.WorkQueueResponse.from_orm(work_queue)
+    return schemas.responses.WorkQueueResponse.model_validate(
+        work_queue, from_attributes=True
+    )
 
 
-@router.get("/{id}")
+@router.get("/{id:uuid}")
 async def read_work_queue(
     work_queue_id: UUID = Path(..., description="The work queue id", alias="id"),
     db: PrefectDBInterface = Depends(provide_database_interface),
@@ -116,28 +123,22 @@ async def read_work_queue(
         raise HTTPException(
             status_code=status.HTTP_404_NOT_FOUND, detail="work queue not found"
         )
-    return schemas.responses.WorkQueueResponse.from_orm(work_queue)
+    return schemas.responses.WorkQueueResponse.model_validate(
+        work_queue, from_attributes=True
+    )
 
 
-@router.post("/{id}/get_runs")
+@router.post("/{id:uuid}/get_runs")
 async def read_work_queue_runs(
-    background_tasks: BackgroundTasks,
+    docket: dependencies.Docket,
     work_queue_id: UUID = Path(..., description="The work queue id", alias="id"),
     limit: int = dependencies.LimitBody(),
-    scheduled_before: DateTimeTZ = Body(
+    scheduled_before: DateTime = Body(
         None,
         description=(
             "Only flow runs scheduled to start before this time will be returned."
         ),
     ),
-    agent_id: Optional[UUID] = Body(
-        None,
-        description=(
-            "An optional unique identifier for the agent making this query. If"
-            " provided, the Prefect REST API will track the last time this agent polled"
-            " the work queue."
-        ),
-    ),
     x_prefect_ui: Optional[bool] = Header(
         default=False,
         description="A header to indicate this request came from the Prefect UI.",
@@ -160,46 +161,25 @@ async def read_work_queue_runs(
     if x_prefect_ui:
         return flow_runs
 
-    background_tasks.add_task(
-        mark_work_queues_ready,
+    await docket.add(mark_work_queues_ready)(
         polled_work_queue_ids=[work_queue_id],
         ready_work_queue_ids=(
             [work_queue_id] if work_queue.status == WorkQueueStatus.NOT_READY else []
         ),
     )
 
-    if agent_id:
-        background_tasks.add_task(
-            _record_agent_poll,
-            work_queue_id=work_queue_id,
-            agent_id=agent_id,
-        )
-
-    background_tasks.add_task(
-        mark_deployments_ready,
+    await docket.add(mark_deployments_ready)(
         work_queue_ids=[work_queue_id],
     )
 
     return flow_runs
 
 
-@db_injector
-async def _record_agent_poll(
-    db: PrefectDBInterface,
-    work_queue_id: UUID,
-    agent_id: UUID,
-):
-    async with db.session_context(begin_transaction=True) as session:
-        await models.agents.record_agent_poll(
-            session=session, agent_id=agent_id, work_queue_id=work_queue_id
-        )
-
-
 @router.post("/filter")
 async def read_work_queues(
     limit: int = dependencies.LimitBody(),
     offset: int = Body(0, ge=0),
-    work_queues: schemas.filters.WorkQueueFilter = None,
+    work_queues: Optional[schemas.filters.WorkQueueFilter] = None,
     db: PrefectDBInterface = Depends(provide_database_interface),
 ) -> List[schemas.responses.WorkQueueResponse]:
     """
@@ -210,14 +190,17 @@ async def read_work_queues(
             session=session, offset=offset, limit=limit, work_queue_filter=work_queues
         )
 
-    return [schemas.responses.WorkQueueResponse.from_orm(wq) for wq in wqs]
+    return [
+        schemas.responses.WorkQueueResponse.model_validate(wq, from_attributes=True)
+        for wq in wqs
+    ]
 
 
-@router.delete("/{id}", status_code=status.HTTP_204_NO_CONTENT)
+@router.delete("/{id:uuid}", status_code=status.HTTP_204_NO_CONTENT)
 async def delete_work_queue(
     work_queue_id: UUID = Path(..., description="The work queue id", alias="id"),
     db: PrefectDBInterface = Depends(provide_database_interface),
-):
+) -> None:
     """
     Delete a work queue by id.
     """
@@ -231,7 +214,7 @@ async def delete_work_queue(
         )
 
 
-@router.get("/{id}/status")
+@router.get("/{id:uuid}/status")
 async def read_work_queue_status(
     work_queue_id: UUID = Path(..., description="The work queue id", alias="id"),
     db: PrefectDBInterface = Depends(provide_database_interface),
