@@ -3,26 +3,25 @@ Routes for interacting with work queue objects.
 """
 
 from typing import TYPE_CHECKING, List, Optional
-from uuid import UUID, uuid4
+from uuid import UUID
 
-import pendulum
 import sqlalchemy as sa
-from prefect._vendor.fastapi import (
-    BackgroundTasks,
+from fastapi import (
     Body,
     Depends,
     HTTPException,
     Path,
     status,
 )
+from packaging.version import Version
 from sqlalchemy.ext.asyncio import AsyncSession
 
 import prefect.server.api.dependencies as dependencies
 import prefect.server.models as models
 import prefect.server.schemas as schemas
+from prefect._internal.uuid7 import uuid7
 from prefect.server.api.validation import validate_job_variable_defaults_for_work_pool
-from prefect.server.database.dependencies import provide_database_interface
-from prefect.server.database.interface import PrefectDBInterface
+from prefect.server.database import PrefectDBInterface, provide_database_interface
 from prefect.server.models.deployments import mark_deployments_ready
 from prefect.server.models.work_queues import (
     emit_work_queue_status_event,
@@ -30,13 +29,14 @@ from prefect.server.models.work_queues import (
 )
 from prefect.server.models.workers import emit_work_pool_status_event
 from prefect.server.schemas.statuses import WorkQueueStatus
-from prefect.server.utilities.schemas import DateTimeTZ
 from prefect.server.utilities.server import PrefectRouter
+from prefect.types import DateTime
+from prefect.types._datetime import now
 
 if TYPE_CHECKING:
     from prefect.server.database.orm_models import ORMWorkQueue
 
-router = PrefectRouter(
+router: PrefectRouter = PrefectRouter(
     prefix="/work_pools",
     tags=["Work Pools"],
 )
@@ -158,18 +158,16 @@ class WorkerLookups:
 async def create_work_pool(
     work_pool: schemas.actions.WorkPoolCreate,
     db: PrefectDBInterface = Depends(provide_database_interface),
+    prefect_client_version: Optional[str] = Depends(
+        dependencies.get_prefect_client_version
+    ),
 ) -> schemas.core.WorkPool:
     """
     Creates a new work pool. If a work pool with the same
     name already exists, an error will be raised.
+
+    For more information, see https://docs.prefect.io/v3/concepts/work-pools.
     """
-
-    if not work_pool.name.lower().strip("' \""):
-        raise HTTPException(
-            status_code=status.HTTP_400_BAD_REQUEST,
-            detail="Work pool name cannot be empty.",
-        )
-
     if work_pool.name.lower().startswith("prefect"):
         raise HTTPException(
             status_code=status.HTTP_403_FORBIDDEN,
@@ -186,13 +184,20 @@ async def create_work_pool(
             )
 
             await emit_work_pool_status_event(
-                event_id=uuid4(),
-                occurred=pendulum.now("UTC"),
+                event_id=uuid7(),
+                occurred=now("UTC"),
                 pre_update_work_pool=None,
                 work_pool=model,
             )
 
-            return schemas.core.WorkPool.from_orm(model)
+            ret = schemas.core.WorkPool.model_validate(model, from_attributes=True)
+            if prefect_client_version and Version(prefect_client_version) <= Version(
+                "3.3.7"
+            ):
+                # Client versions 3.3.7 and below do not support the default_result_storage_block_id field and will error
+                # when receiving it.
+                del ret.storage_configuration.default_result_storage_block_id
+            return ret
 
     except sa.exc.IntegrityError:
         raise HTTPException(
@@ -206,6 +211,9 @@ async def read_work_pool(
     work_pool_name: str = Path(..., description="The work pool name", alias="name"),
     worker_lookups: WorkerLookups = Depends(WorkerLookups),
     db: PrefectDBInterface = Depends(provide_database_interface),
+    prefect_client_version: Optional[str] = Depends(
+        dependencies.get_prefect_client_version
+    ),
 ) -> schemas.core.WorkPool:
     """
     Read a work pool by name
@@ -218,7 +226,18 @@ async def read_work_pool(
         orm_work_pool = await models.workers.read_work_pool(
             session=session, work_pool_id=work_pool_id
         )
-        return schemas.core.WorkPool.from_orm(orm_work_pool)
+        work_pool = schemas.core.WorkPool.model_validate(
+            orm_work_pool, from_attributes=True
+        )
+
+        if prefect_client_version and Version(prefect_client_version) <= Version(
+            "3.3.7"
+        ):
+            # Client versions 3.3.7 and below do not support the default_result_storage_block_id field and will error
+            # when receiving it.
+            del work_pool.storage_configuration.default_result_storage_block_id
+
+        return work_pool
 
 
 @router.post("/filter")
@@ -226,8 +245,10 @@ async def read_work_pools(
     work_pools: Optional[schemas.filters.WorkPoolFilter] = None,
     limit: int = dependencies.LimitBody(),
     offset: int = Body(0, ge=0),
-    worker_lookups: WorkerLookups = Depends(WorkerLookups),
     db: PrefectDBInterface = Depends(provide_database_interface),
+    prefect_client_version: Optional[str] = Depends(
+        dependencies.get_prefect_client_version
+    ),
 ) -> List[schemas.core.WorkPool]:
     """
     Read multiple work pools
@@ -239,7 +260,18 @@ async def read_work_pools(
             offset=offset,
             limit=limit,
         )
-        return [schemas.core.WorkPool.from_orm(w) for w in orm_work_pools]
+        ret = [
+            schemas.core.WorkPool.model_validate(w, from_attributes=True)
+            for w in orm_work_pools
+        ]
+        if prefect_client_version and Version(prefect_client_version) <= Version(
+            "3.3.7"
+        ):
+            # Client versions 3.3.7 and below do not support the default_result_storage_block_id field and will error
+            # when receiving it.
+            for work_pool in ret:
+                del work_pool.storage_configuration.default_result_storage_block_id
+        return ret
 
 
 @router.post("/count")
@@ -262,13 +294,13 @@ async def update_work_pool(
     work_pool_name: str = Path(..., description="The work pool name", alias="name"),
     worker_lookups: WorkerLookups = Depends(WorkerLookups),
     db: PrefectDBInterface = Depends(provide_database_interface),
-):
+) -> None:
     """
     Update a work pool
     """
 
     # Reserved pools can only updated pause / concurrency
-    update_values = work_pool.dict(exclude_unset=True)
+    update_values = work_pool.model_dump(exclude_unset=True)
     if work_pool_name.lower().startswith("prefect") and (
         set(update_values).difference({"is_paused", "concurrency_limit"})
     ):
@@ -297,7 +329,7 @@ async def delete_work_pool(
     work_pool_name: str = Path(..., description="The work pool name", alias="name"),
     worker_lookups: WorkerLookups = Depends(WorkerLookups),
     db: PrefectDBInterface = Depends(provide_database_interface),
-):
+) -> None:
     """
     Delete a work pool
     """
@@ -323,15 +355,15 @@ async def delete_work_pool(
 
 @router.post("/{name}/get_scheduled_flow_runs")
 async def get_scheduled_flow_runs(
-    background_tasks: BackgroundTasks,
+    docket: dependencies.Docket,
     work_pool_name: str = Path(..., description="The work pool name", alias="name"),
     work_queue_names: List[str] = Body(
         None, description="The names of work pool queues"
     ),
-    scheduled_before: DateTimeTZ = Body(
+    scheduled_before: DateTime = Body(
         None, description="The maximum time to look for scheduled flow runs"
     ),
-    scheduled_after: DateTimeTZ = Body(
+    scheduled_after: DateTime = Body(
         None, description="The minimum time to look for scheduled flow runs"
     ),
     limit: int = dependencies.LimitBody(),
@@ -376,8 +408,7 @@ async def get_scheduled_flow_runs(
             limit=limit,
         )
 
-    background_tasks.add_task(
-        mark_work_queues_ready,
+    await docket.add(mark_work_queues_ready)(
         polled_work_queue_ids=[
             wq.id for wq in work_queues if wq.status != WorkQueueStatus.NOT_READY
         ],
@@ -386,8 +417,7 @@ async def get_scheduled_flow_runs(
         ],
     )
 
-    background_tasks.add_task(
-        mark_deployments_ready,
+    await docket.add(mark_deployments_ready)(
         work_queue_ids=[wq.id for wq in work_queues],
     )
 
@@ -413,6 +443,8 @@ async def create_work_queue(
     """
     Creates a new work pool queue. If a work pool queue with the same
     name already exists, an error will be raised.
+
+    For more information, see https://docs.prefect.io/v3/concepts/work-pools#work-queues.
     """
 
     try:
@@ -436,7 +468,9 @@ async def create_work_queue(
             ),
         )
 
-    return schemas.responses.WorkQueueResponse.from_orm(model)
+    return schemas.responses.WorkQueueResponse.model_validate(
+        model, from_attributes=True
+    )
 
 
 @router.get("/{work_pool_name}/queues/{name}")
@@ -463,7 +497,9 @@ async def read_work_queue(
             session=session, work_queue_id=work_queue_id
         )
 
-    return schemas.responses.WorkQueueResponse.from_orm(model)
+    return schemas.responses.WorkQueueResponse.model_validate(
+        model, from_attributes=True
+    )
 
 
 @router.post("/{work_pool_name}/queues/filter")
@@ -491,7 +527,10 @@ async def read_work_queues(
             offset=offset,
         )
 
-    return [schemas.responses.WorkQueueResponse.from_orm(wq) for wq in wqs]
+    return [
+        schemas.responses.WorkQueueResponse.model_validate(wq, from_attributes=True)
+        for wq in wqs
+    ]
 
 
 @router.patch("/{work_pool_name}/queues/{name}", status_code=status.HTTP_204_NO_CONTENT)
@@ -503,7 +542,7 @@ async def update_work_queue(
     ),
     worker_lookups: WorkerLookups = Depends(WorkerLookups),
     db: PrefectDBInterface = Depends(provide_database_interface),
-):
+) -> None:
     """
     Update a work pool queue
     """
@@ -533,7 +572,7 @@ async def delete_work_queue(
     ),
     worker_lookups: WorkerLookups = Depends(WorkerLookups),
     db: PrefectDBInterface = Depends(provide_database_interface),
-):
+) -> None:
     """
     Delete a work pool queue
     """
@@ -571,7 +610,7 @@ async def worker_heartbeat(
     ),
     worker_lookups: WorkerLookups = Depends(WorkerLookups),
     db: PrefectDBInterface = Depends(provide_database_interface),
-):
+) -> None:
     async with db.session_context(begin_transaction=True) as session:
         work_pool = await models.workers.read_work_pool_by_name(
             session=session,
@@ -604,7 +643,7 @@ async def worker_heartbeat(
 @router.post("/{work_pool_name}/workers/filter")
 async def read_workers(
     work_pool_name: str = Path(..., description="The work pool name"),
-    workers: schemas.filters.WorkerFilter = None,
+    workers: Optional[schemas.filters.WorkerFilter] = None,
     limit: int = dependencies.LimitBody(),
     offset: int = Body(0, ge=0),
     worker_lookups: WorkerLookups = Depends(WorkerLookups),
@@ -636,7 +675,7 @@ async def delete_worker(
     ),
     worker_lookups: WorkerLookups = Depends(WorkerLookups),
     db: PrefectDBInterface = Depends(provide_database_interface),
-):
+) -> None:
     """
     Delete a work pool's worker
     """
