@@ -1,49 +1,57 @@
+from __future__ import annotations
+
 import sqlite3
+import ssl
 import traceback
 from abc import ABC, abstractmethod
 from asyncio import AbstractEventLoop, get_running_loop
-from contextlib import asynccontextmanager
+from collections.abc import AsyncGenerator, Hashable
+from contextlib import AbstractAsyncContextManager, asynccontextmanager
 from contextvars import ContextVar
 from functools import partial
-from typing import Dict, Hashable, Optional, Tuple
+from typing import Any, Optional
 
 import sqlalchemy as sa
+from sqlalchemy import AdaptedConnection, event
+from sqlalchemy.dialects.sqlite import aiosqlite
+from sqlalchemy.engine.interfaces import DBAPIConnection
+from sqlalchemy.ext.asyncio import (
+    AsyncConnection,
+    AsyncEngine,
+    AsyncSession,
+    AsyncSessionTransaction,
+    create_async_engine,
+)
+from sqlalchemy.pool import ConnectionPoolEntry
+from typing_extensions import TypeAlias
 
-try:
-    from sqlalchemy import AdaptedConnection
-    from sqlalchemy.pool import ConnectionPoolEntry
-except ImportError:
-    # SQLAlchemy 1.4 equivalents
-    from sqlalchemy.pool import _ConnectionFairy as AdaptedConnection
-    from sqlalchemy.pool.base import _ConnectionRecord as ConnectionPoolEntry
-
-from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, create_async_engine
-from typing_extensions import Literal
-
+from prefect._internal.observability import configure_logfire
 from prefect.settings import (
     PREFECT_API_DATABASE_CONNECTION_TIMEOUT,
     PREFECT_API_DATABASE_ECHO,
     PREFECT_API_DATABASE_TIMEOUT,
-    PREFECT_SQLALCHEMY_MAX_OVERFLOW,
-    PREFECT_SQLALCHEMY_POOL_SIZE,
-    PREFECT_UNIT_TEST_MODE,
+    PREFECT_TESTING_UNIT_TEST_MODE,
+    get_current_settings,
 )
 from prefect.utilities.asyncutils import add_event_loop_shutdown_callback
 
+logfire: Any | None = configure_logfire()
+
 SQLITE_BEGIN_MODE: ContextVar[Optional[str]] = ContextVar(  # novm
     "SQLITE_BEGIN_MODE", default=None
 )
 
-ENGINES: Dict[Tuple[AbstractEventLoop, str, bool, float], AsyncEngine] = {}
+_EngineCacheKey: TypeAlias = tuple[AbstractEventLoop, str, bool, Optional[float]]
+ENGINES: dict[_EngineCacheKey, AsyncEngine] = {}
 
 
 class ConnectionTracker:
     """A test utility which tracks the connections given out by a connection pool, to
     make it easy to see which connections are currently checked out and open."""
 
-    all_connections: Dict[AdaptedConnection, str]
-    open_connections: Dict[AdaptedConnection, str]
-    left_field_closes: Dict[AdaptedConnection, str]
+    all_connections: dict[AdaptedConnection, list[str]]
+    open_connections: dict[AdaptedConnection, list[str]]
+    left_field_closes: dict[AdaptedConnection, list[str]]
     connects: int
     closes: int
     active: bool
@@ -56,16 +64,16 @@ class ConnectionTracker:
         self.connects = 0
         self.closes = 0
 
-    def track_pool(self, pool: sa.pool.Pool):
-        sa.event.listen(pool, "connect", self.on_connect)
-        sa.event.listen(pool, "close", self.on_close)
-        sa.event.listen(pool, "close_detached", self.on_close_detached)
+    def track_pool(self, pool: sa.pool.Pool) -> None:
+        event.listen(pool, "connect", self.on_connect)
+        event.listen(pool, "close", self.on_close)
+        event.listen(pool, "close_detached", self.on_close_detached)
 
     def on_connect(
         self,
         adapted_connection: AdaptedConnection,
         connection_record: ConnectionPoolEntry,
-    ):
+    ) -> None:
         self.all_connections[adapted_connection] = traceback.format_stack()
         self.open_connections[adapted_connection] = traceback.format_stack()
         self.connects += 1
@@ -74,7 +82,7 @@ class ConnectionTracker:
         self,
         adapted_connection: AdaptedConnection,
         connection_record: ConnectionPoolEntry,
-    ):
+    ) -> None:
         try:
             del self.open_connections[adapted_connection]
         except KeyError:
@@ -84,14 +92,14 @@ class ConnectionTracker:
     def on_close_detached(
         self,
         adapted_connection: AdaptedConnection,
-    ):
+    ) -> None:
         try:
             del self.open_connections[adapted_connection]
         except KeyError:
             self.left_field_closes[adapted_connection] = traceback.format_stack()
         self.closes += 1
 
-    def clear(self):
+    def clear(self) -> None:
         self.all_connections.clear()
         self.open_connections.clear()
         self.left_field_closes.clear()
@@ -113,26 +121,43 @@ class BaseDatabaseConfiguration(ABC):
     def __init__(
         self,
         connection_url: str,
-        echo: bool = None,
-        timeout: float = None,
-        connection_timeout: float = None,
-        sqlalchemy_pool_size: int = None,
-        sqlalchemy_max_overflow: int = None,
-    ):
+        echo: Optional[bool] = None,
+        timeout: Optional[float] = None,
+        connection_timeout: Optional[float] = None,
+        sqlalchemy_pool_size: Optional[int] = None,
+        sqlalchemy_max_overflow: Optional[int] = None,
+        connection_app_name: Optional[str] = None,
+        statement_cache_size: Optional[int] = None,
+        prepared_statement_cache_size: Optional[int] = None,
+    ) -> None:
         self.connection_url = connection_url
-        self.echo = echo or PREFECT_API_DATABASE_ECHO.value()
-        self.timeout = timeout or PREFECT_API_DATABASE_TIMEOUT.value()
-        self.connection_timeout = (
+        self.echo: bool = echo or PREFECT_API_DATABASE_ECHO.value()
+        self.timeout: Optional[float] = timeout or PREFECT_API_DATABASE_TIMEOUT.value()
+        self.connection_timeout: Optional[float] = (
             connection_timeout or PREFECT_API_DATABASE_CONNECTION_TIMEOUT.value()
         )
-        self.sqlalchemy_pool_size = (
-            sqlalchemy_pool_size or PREFECT_SQLALCHEMY_POOL_SIZE.value()
+        self.sqlalchemy_pool_size: Optional[int] = (
+            sqlalchemy_pool_size
+            or get_current_settings().server.database.sqlalchemy.pool_size
         )
-        self.sqlalchemy_max_overflow = (
-            sqlalchemy_max_overflow or PREFECT_SQLALCHEMY_MAX_OVERFLOW.value()
+        self.sqlalchemy_max_overflow: Optional[int] = (
+            sqlalchemy_max_overflow
+            or get_current_settings().server.database.sqlalchemy.max_overflow
+        )
+        self.connection_app_name: Optional[str] = (
+            connection_app_name
+            or get_current_settings().server.database.sqlalchemy.connect_args.application_name
+        )
+        self.statement_cache_size: Optional[int] = (
+            statement_cache_size
+            or get_current_settings().server.database.sqlalchemy.connect_args.statement_cache_size
+        )
+        self.prepared_statement_cache_size: Optional[int] = (
+            prepared_statement_cache_size
+            or get_current_settings().server.database.sqlalchemy.connect_args.prepared_statement_cache_size
         )
 
-    def _unique_key(self) -> Tuple[Hashable, ...]:
+    def unique_key(self) -> tuple[Hashable, ...]:
         """
         Returns a key used to determine whether to instantiate a new DB interface.
         """
@@ -149,11 +174,15 @@ class BaseDatabaseConfiguration(ABC):
         """
 
     @abstractmethod
-    async def create_db(self, connection, base_metadata):
+    async def create_db(
+        self, connection: AsyncConnection, base_metadata: sa.MetaData
+    ) -> None:
         """Create the database"""
 
     @abstractmethod
-    async def drop_db(self, connection, base_metadata):
+    async def drop_db(
+        self, connection: AsyncConnection, base_metadata: sa.MetaData
+    ) -> None:
         """Drop the database"""
 
     @abstractmethod
@@ -161,9 +190,9 @@ class BaseDatabaseConfiguration(ABC):
         """Returns true if database is run in memory"""
 
     @abstractmethod
-    async def begin_transaction(
+    def begin_transaction(
         self, session: AsyncSession, with_for_update: bool = False
-    ):
+    ) -> AbstractAsyncContextManager[AsyncSessionTransaction]:
         """Enter a transaction for a session"""
         pass
 
@@ -193,9 +222,12 @@ class AsyncPostgresConfiguration(BaseDatabaseConfiguration):
             self.timeout,
         )
         if cache_key not in ENGINES:
-            # apply database timeout
-            kwargs = dict()
-            connect_args = dict()
+            kwargs: dict[str, Any] = (
+                get_current_settings().server.database.sqlalchemy.model_dump(
+                    mode="json", exclude={"connect_args"}
+                )
+            )
+            connect_args: dict[str, Any] = {}
 
             if self.timeout is not None:
                 connect_args["command_timeout"] = self.timeout
@@ -203,8 +235,43 @@ class AsyncPostgresConfiguration(BaseDatabaseConfiguration):
             if self.connection_timeout is not None:
                 connect_args["timeout"] = self.connection_timeout
 
+            if self.statement_cache_size is not None:
+                connect_args["statement_cache_size"] = self.statement_cache_size
+
+            if self.prepared_statement_cache_size is not None:
+                connect_args["prepared_statement_cache_size"] = (
+                    self.prepared_statement_cache_size
+                )
+
+            if self.connection_app_name is not None:
+                connect_args["server_settings"] = dict(
+                    application_name=self.connection_app_name
+                )
+
+            if get_current_settings().server.database.sqlalchemy.connect_args.tls.enabled:
+                tls_config = (
+                    get_current_settings().server.database.sqlalchemy.connect_args.tls
+                )
+
+                pg_ctx = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
+
+                if tls_config.ca_file:
+                    pg_ctx = ssl.create_default_context(
+                        purpose=ssl.Purpose.SERVER_AUTH, cafile=tls_config.ca_file
+                    )
+
+                pg_ctx.minimum_version = ssl.TLSVersion.TLSv1_2
+
+                if tls_config.cert_file and tls_config.key_file:
+                    pg_ctx.load_cert_chain(
+                        certfile=tls_config.cert_file, keyfile=tls_config.key_file
+                    )
+
+                pg_ctx.check_hostname = tls_config.check_hostname
+                pg_ctx.verify_mode = ssl.CERT_REQUIRED
+                connect_args["ssl"] = pg_ctx
+
             if connect_args:
-                connect_args["server_settings"] = {"jit": "off"}
                 kwargs["connect_args"] = connect_args
 
             if self.sqlalchemy_pool_size is not None:
@@ -227,6 +294,9 @@ class AsyncPostgresConfiguration(BaseDatabaseConfiguration):
                 **kwargs,
             )
 
+            if logfire:
+                logfire.instrument_sqlalchemy(engine)  # pyright: ignore
+
             if TRACKER.active:
                 TRACKER.track_pool(engine.pool)
 
@@ -234,7 +304,7 @@ class AsyncPostgresConfiguration(BaseDatabaseConfiguration):
             await self.schedule_engine_disposal(cache_key)
         return ENGINES[cache_key]
 
-    async def schedule_engine_disposal(self, cache_key):
+    async def schedule_engine_disposal(self, cache_key: _EngineCacheKey) -> None:
         """
         Dispose of an engine once the event loop is closing.
 
@@ -251,7 +321,7 @@ class AsyncPostgresConfiguration(BaseDatabaseConfiguration):
         encountered should be encouraged to use a standalone server.
         """
 
-        async def dispose_engine(cache_key):
+        async def dispose_engine(cache_key: _EngineCacheKey) -> None:
             engine = ENGINES.pop(cache_key, None)
             if engine:
                 await engine.dispose()
@@ -270,23 +340,27 @@ class AsyncPostgresConfiguration(BaseDatabaseConfiguration):
     @asynccontextmanager
     async def begin_transaction(
         self, session: AsyncSession, with_for_update: bool = False
-    ):
+    ) -> AsyncGenerator[AsyncSessionTransaction, None]:
         # `with_for_update` is for SQLite only. For Postgres, lock the row on read
         # for update instead.
         async with session.begin() as transaction:
             yield transaction
 
-    async def create_db(self, connection, base_metadata):
+    async def create_db(
+        self, connection: AsyncConnection, base_metadata: sa.MetaData
+    ) -> None:
         """Create the database"""
 
         await connection.run_sync(base_metadata.create_all)
 
-    async def drop_db(self, connection, base_metadata):
+    async def drop_db(
+        self, connection: AsyncConnection, base_metadata: sa.MetaData
+    ) -> None:
         """Drop the database"""
 
         await connection.run_sync(base_metadata.drop_all)
 
-    def is_inmemory(self) -> Literal[False]:
+    def is_inmemory(self) -> bool:
         """Returns true if database is run in memory"""
 
         return False
@@ -317,16 +391,11 @@ class AioSqliteConfiguration(BaseDatabaseConfiguration):
                 f"{sqlite3.sqlite_version}"
             )
 
-        kwargs = {}
+        kwargs: dict[str, Any] = dict()
 
         loop = get_running_loop()
 
-        cache_key = (
-            loop,
-            self.connection_url,
-            self.echo,
-            self.timeout,
-        )
+        cache_key = (loop, self.connection_url, self.echo, self.timeout)
         if cache_key not in ENGINES:
             # apply database timeout
             if self.timeout is not None:
@@ -341,11 +410,19 @@ class AioSqliteConfiguration(BaseDatabaseConfiguration):
             # ensure a long-lasting pool is used with in-memory databases
             # because they disappear when the last connection closes
             if ":memory:" in self.connection_url:
-                kwargs.update(poolclass=sa.pool.SingletonThreadPool)
+                kwargs.update(
+                    poolclass=sa.pool.AsyncAdaptedQueuePool,
+                    pool_size=1,
+                    max_overflow=0,
+                    pool_recycle=-1,
+                )
 
             engine = create_async_engine(self.connection_url, echo=self.echo, **kwargs)
-            sa.event.listen(engine.sync_engine, "connect", self.setup_sqlite)
-            sa.event.listen(engine.sync_engine, "begin", self.begin_sqlite_stmt)
+            event.listen(engine.sync_engine, "connect", self.setup_sqlite)
+            event.listen(engine.sync_engine, "begin", self.begin_sqlite_stmt)
+
+            if logfire:
+                logfire.instrument_sqlalchemy(engine)  # pyright: ignore
 
             if TRACKER.active:
                 TRACKER.track_pool(engine.pool)
@@ -354,7 +431,7 @@ class AioSqliteConfiguration(BaseDatabaseConfiguration):
             await self.schedule_engine_disposal(cache_key)
         return ENGINES[cache_key]
 
-    async def schedule_engine_disposal(self, cache_key):
+    async def schedule_engine_disposal(self, cache_key: _EngineCacheKey) -> None:
         """
         Dispose of an engine once the event loop is closing.
 
@@ -371,19 +448,20 @@ class AioSqliteConfiguration(BaseDatabaseConfiguration):
         encountered should be encouraged to use a standalone server.
         """
 
-        async def dispose_engine(cache_key):
+        async def dispose_engine(cache_key: _EngineCacheKey) -> None:
             engine = ENGINES.pop(cache_key, None)
             if engine:
                 await engine.dispose()
 
         await add_event_loop_shutdown_callback(partial(dispose_engine, cache_key))
 
-    def setup_sqlite(self, conn, record):
+    def setup_sqlite(self, conn: DBAPIConnection, record: ConnectionPoolEntry) -> None:
         """Issue PRAGMA statements to SQLITE on connect. PRAGMAs only last for the
         duration of the connection. See https://www.sqlite.org/pragma.html for more info.
         """
         # workaround sqlite transaction behavior
-        self.begin_sqlite_conn(conn, record)
+        if isinstance(conn, aiosqlite.AsyncAdapt_aiosqlite_connection):
+            self.begin_sqlite_conn(conn)
 
         cursor = conn.cursor()
 
@@ -416,21 +494,28 @@ class AioSqliteConfiguration(BaseDatabaseConfiguration):
         # before returning and raising an error
         # setting the value very high allows for more 'concurrency'
         # without running into errors, but may result in slow api calls
-        if PREFECT_UNIT_TEST_MODE.value() is True:
+        if PREFECT_TESTING_UNIT_TEST_MODE.value() is True:
             cursor.execute("PRAGMA busy_timeout = 5000;")  # 5s
         else:
             cursor.execute("PRAGMA busy_timeout = 60000;")  # 60s
 
+        # `PRAGMA temp_store = memory;` moves temporary tables from disk into RAM
+        # this supposedly speeds up reads, but it seems to actually
+        # decrease overall performance, see https://github.com/PrefectHQ/prefect/pull/14812
+        # cursor.execute("PRAGMA temp_store = memory;")
+
         cursor.close()
 
-    def begin_sqlite_conn(self, conn, record):
+    def begin_sqlite_conn(
+        self, conn: aiosqlite.AsyncAdapt_aiosqlite_connection
+    ) -> None:
         # disable pysqlite's emitting of the BEGIN statement entirely.
         # also stops it from emitting COMMIT before any DDL.
         # requires `begin_sqlite_stmt`
         # see https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#serializable-isolation-savepoints-transactional-ddl
         conn.isolation_level = None
 
-    def begin_sqlite_stmt(self, conn):
+    def begin_sqlite_stmt(self, conn: sa.Connection) -> None:
         # emit our own BEGIN
         # requires `begin_sqlite_conn`
         # see https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#serializable-isolation-savepoints-transactional-ddl
@@ -445,7 +530,7 @@ class AioSqliteConfiguration(BaseDatabaseConfiguration):
     @asynccontextmanager
     async def begin_transaction(
         self, session: AsyncSession, with_for_update: bool = False
-    ):
+    ) -> AsyncGenerator[AsyncSessionTransaction, None]:
         token = SQLITE_BEGIN_MODE.set("IMMEDIATE" if with_for_update else "DEFERRED")
 
         try:
@@ -463,17 +548,21 @@ class AioSqliteConfiguration(BaseDatabaseConfiguration):
         """
         return AsyncSession(engine, expire_on_commit=False)
 
-    async def create_db(self, connection, base_metadata):
+    async def create_db(
+        self, connection: AsyncConnection, base_metadata: sa.MetaData
+    ) -> None:
         """Create the database"""
 
         await connection.run_sync(base_metadata.create_all)
 
-    async def drop_db(self, connection, base_metadata):
+    async def drop_db(
+        self, connection: AsyncConnection, base_metadata: sa.MetaData
+    ) -> None:
         """Drop the database"""
 
         await connection.run_sync(base_metadata.drop_all)
 
-    def is_inmemory(self):
+    def is_inmemory(self) -> bool:
         """Returns true if database is run in memory"""
 
         return ":memory:" in self.connection_url or "mode=memory" in self.connection_url
