@@ -1,72 +1,77 @@
-from typing import List, Optional, Union
+from typing import List, Optional, Sequence, Union
 from uuid import UUID
 
 import sqlalchemy as sa
+from sqlalchemy.ext.asyncio import AsyncSession
+from sqlalchemy.sql.elements import ColumnElement
 
 import prefect.server.schemas as schemas
-from prefect.server.database.dependencies import inject_db
-from prefect.server.database.interface import PrefectDBInterface
+from prefect.server.database import PrefectDBInterface, db_injector, orm_models
+from prefect.server.utilities.database import greatest, least
+from prefect.settings import get_current_settings
 
 
-def greatest(db, clamped_value, sql_value):
-    # Determine the greatest value based on the database type
-    if db.dialect.name == "sqlite":
-        # `sa.func.greatest` isn't available in SQLite, fallback to using a
-        # `case` statement.
-        return sa.case((clamped_value > sql_value, clamped_value), else_=sql_value)
-    else:
-        return sa.func.greatest(clamped_value, sql_value)
-
-
-def seconds_ago(db, field) -> float:
-    if db.dialect.name == "sqlite":
-        # `sa.func.timezone` isn't available in SQLite, fallback to using
-        # `julianday` .
-        return (sa.func.julianday("now") - sa.func.julianday(field)) * 86400.0
-    else:
-        return sa.func.extract(
-            "epoch",
-            sa.func.timezone("UTC", sa.func.now()) - sa.func.timezone("UTC", field),
-        ).cast(sa.Float)
-
-
-def active_slots_after_decay(db: PrefectDBInterface):
+def active_slots_after_decay(db: PrefectDBInterface) -> ColumnElement[float]:
     # Active slots will decay at a rate of `slot_decay_per_second` per second.
     return greatest(
-        db,
         0,
         db.ConcurrencyLimitV2.active_slots
         - sa.func.floor(
             db.ConcurrencyLimitV2.slot_decay_per_second
-            * seconds_ago(db, db.ConcurrencyLimitV2.updated)
+            * sa.func.date_diff_seconds(db.ConcurrencyLimitV2.updated)
         ),
     )
 
 
-def denied_slots_after_decay(db: PrefectDBInterface):
-    # Denied slots decay at a rate of `slot_decay_per_second` per second if it's
-    # greater than 0, otherwise it decays at a rate of `avg_slot_occupancy_seconds`.
-    # The combination of `denied_slots` and `slot_decay_per_second` /
-    # `avg_slot_occupancy_seconds` is used to by the API to give a best guess at
-    # when slots will be available again.
+def denied_slots_after_decay(db: PrefectDBInterface) -> ColumnElement[float]:
+    """
+    Calculate denied_slots after applying decay.
+
+    Denied slots decay at a rate of `slot_decay_per_second` per second if it's
+    greater than 0 (rate limits), otherwise for concurrency limits it decays at
+    a rate based on clamped `avg_slot_occupancy_seconds`.
+
+    The clamping matches the retry-after calculation to prevent denied_slots from
+    accumulating when clients retry faster than the unclamped decay rate.
+    """
+    settings = get_current_settings()
+
+    # Determine max_wait based on limit name prefix
+    max_wait_for_limit = sa.case(
+        (
+            db.ConcurrencyLimitV2.name.like("tag:%"),
+            sa.literal(settings.server.tasks.tag_concurrency_slot_wait_seconds),
+        ),
+        else_=sa.literal(
+            settings.server.concurrency.maximum_concurrency_slot_wait_seconds
+        ),
+    )
+
+    # Clamp avg_slot_occupancy_seconds with minimum bound to prevent division by zero
+    clamped_occupancy = greatest(
+        sa.literal(MINIMUM_OCCUPANCY_SECONDS_PER_SLOT),
+        least(
+            sa.cast(db.ConcurrencyLimitV2.avg_slot_occupancy_seconds, sa.Float),
+            max_wait_for_limit,
+        ),
+    )
+
+    # Calculate decay rate: use slot_decay_per_second for rate limits,
+    # use 1/clamped_occupancy for concurrency limits
+    decay_rate_per_second = sa.case(
+        (
+            db.ConcurrencyLimitV2.slot_decay_per_second > 0.0,
+            db.ConcurrencyLimitV2.slot_decay_per_second,  # Rate limits - no clamping
+        ),
+        else_=(1.0 / clamped_occupancy),  # Concurrency limits - use clamped value
+    )
+
     return greatest(
-        db,
         0,
         db.ConcurrencyLimitV2.denied_slots
         - sa.func.floor(
-            sa.case(
-                (
-                    db.ConcurrencyLimitV2.slot_decay_per_second > 0.0,
-                    db.ConcurrencyLimitV2.slot_decay_per_second,
-                ),
-                else_=(
-                    1.0
-                    / sa.cast(
-                        db.ConcurrencyLimitV2.avg_slot_occupancy_seconds, sa.Float
-                    )
-                ),
-            )
-            * seconds_ago(db, db.ConcurrencyLimitV2.updated)
+            decay_rate_per_second
+            * sa.func.date_diff_seconds(db.ConcurrencyLimitV2.updated)
         ),
     )
 
@@ -80,15 +85,15 @@ OCCUPANCY_SAMPLES_MULTIPLIER = 2
 MINIMUM_OCCUPANCY_SECONDS_PER_SLOT = 0.1
 
 
-@inject_db
+@db_injector
 async def create_concurrency_limit(
-    session: sa.orm.Session,
     db: PrefectDBInterface,
+    session: AsyncSession,
     concurrency_limit: Union[
         schemas.actions.ConcurrencyLimitV2Create, schemas.core.ConcurrencyLimitV2
     ],
-):
-    model = db.ConcurrencyLimitV2(**concurrency_limit.dict())
+) -> orm_models.ConcurrencyLimitV2:
+    model = db.ConcurrencyLimitV2(**concurrency_limit.model_dump())
 
     session.add(model)
     await session.flush()
@@ -96,13 +101,13 @@ async def create_concurrency_limit(
     return model
 
 
-@inject_db
+@db_injector
 async def read_concurrency_limit(
-    session: sa.orm.Session,
     db: PrefectDBInterface,
+    session: AsyncSession,
     concurrency_limit_id: Optional[UUID] = None,
     name: Optional[str] = None,
-):
+) -> Union[orm_models.ConcurrencyLimitV2, None]:
     if not concurrency_limit_id and not name:
         raise ValueError("Must provide either concurrency_limit_id or name")
 
@@ -116,13 +121,13 @@ async def read_concurrency_limit(
     return result.scalar()
 
 
-@inject_db
+@db_injector
 async def read_all_concurrency_limits(
-    session: sa.orm.Session,
     db: PrefectDBInterface,
+    session: AsyncSession,
     limit: int,
     offset: int,
-):
+) -> Sequence[orm_models.ConcurrencyLimitV2]:
     query = sa.select(db.ConcurrencyLimitV2).order_by(db.ConcurrencyLimitV2.name)
 
     if offset is not None:
@@ -134,10 +139,10 @@ async def read_all_concurrency_limits(
     return result.scalars().unique().all()
 
 
-@inject_db
+@db_injector
 async def update_concurrency_limit(
-    session: sa.orm.Session,
     db: PrefectDBInterface,
+    session: AsyncSession,
     concurrency_limit: schemas.actions.ConcurrencyLimitV2Update,
     concurrency_limit_id: Optional[UUID] = None,
     name: Optional[str] = None,
@@ -160,16 +165,16 @@ async def update_concurrency_limit(
     result = await session.execute(
         sa.update(db.ConcurrencyLimitV2)
         .where(where)
-        .values(**concurrency_limit.dict(exclude_unset=True))
+        .values(**concurrency_limit.model_dump(exclude_unset=True))
     )
 
     return result.rowcount > 0
 
 
-@inject_db
+@db_injector
 async def delete_concurrency_limit(
-    session: sa.orm.Session,
     db: PrefectDBInterface,
+    session: AsyncSession,
     concurrency_limit_id: Optional[UUID] = None,
     name: Optional[str] = None,
 ) -> bool:
@@ -187,42 +192,25 @@ async def delete_concurrency_limit(
     return result.rowcount > 0
 
 
-@inject_db
-async def bulk_read_or_create_concurrency_limits(
-    session: sa.orm.Session,
+@db_injector
+async def bulk_read_concurrency_limits(
     db: PrefectDBInterface,
+    session: AsyncSession,
     names: List[str],
-):
+) -> List[orm_models.ConcurrencyLimitV2]:
     # Get all existing concurrency limits in `names`.
     existing_query = sa.select(db.ConcurrencyLimitV2).where(
         db.ConcurrencyLimitV2.name.in_(names)
     )
     existing_limits = list((await session.execute(existing_query)).scalars().all())
 
-    # If any limits aren't present in the database, create them as inactive.
-    missing_names = set(names) - {str(limit.name) for limit in existing_limits}
-
-    if missing_names:
-        new_limits = [
-            db.ConcurrencyLimitV2(
-                **schemas.core.ConcurrencyLimitV2(
-                    name=name, limit=1, active=False
-                ).dict()
-            )
-            for name in missing_names
-        ]
-        session.add_all(new_limits)
-        await session.flush()
-
-        existing_limits += new_limits
-
     return existing_limits
 
 
-@inject_db
+@db_injector
 async def bulk_increment_active_slots(
-    session: sa.orm.Session,
     db: PrefectDBInterface,
+    session: AsyncSession,
     concurrency_limit_ids: List[UUID],
     slots: int,
 ) -> bool:
@@ -248,10 +236,10 @@ async def bulk_increment_active_slots(
     return result.rowcount == len(concurrency_limit_ids)
 
 
-@inject_db
+@db_injector
 async def bulk_decrement_active_slots(
-    session: sa.orm.Session,
     db: PrefectDBInterface,
+    session: AsyncSession,
     concurrency_limit_ids: List[UUID],
     slots: int,
     occupancy_seconds: Optional[float] = None,
@@ -266,10 +254,7 @@ async def bulk_decrement_active_slots(
         )
         .values(
             active_slots=sa.case(
-                (
-                    active_slots_after_decay(db) - slots < 0,
-                    0,
-                ),
+                (active_slots_after_decay(db) - slots < 0, 0),
                 else_=active_slots_after_decay(db) - slots,
             ),
             denied_slots=denied_slots_after_decay(db),
@@ -299,13 +284,13 @@ async def bulk_decrement_active_slots(
     return result.rowcount == len(concurrency_limit_ids)
 
 
-@inject_db
+@db_injector
 async def bulk_update_denied_slots(
-    session: sa.orm.Session,
+    db: PrefectDBInterface,
+    session: AsyncSession,
     concurrency_limit_ids: List[UUID],
     slots: int,
-    db: PrefectDBInterface,
-):
+) -> bool:
     query = (
         sa.update(db.ConcurrencyLimitV2)
         .where(
