@@ -5,51 +5,67 @@ Utilities for working with Python callables.
 import ast
 import importlib.util
 import inspect
+import warnings
+from collections import OrderedDict
+from collections.abc import Iterable
 from functools import partial
+from logging import Logger
 from pathlib import Path
-from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple
-
-import cloudpickle
-
-from prefect._internal.pydantic import HAS_PYDANTIC_V2
-from prefect._internal.pydantic.v1_schema import has_v1_type_as_param
-
-if HAS_PYDANTIC_V2:
-    import pydantic.v1 as pydantic
-
-    from prefect._internal.pydantic.v2_schema import (
-        create_v2_schema,
-        process_v2_params,
-    )
-else:
-    import pydantic
+from typing import Any, Callable, Optional, Union, cast
 
+import cloudpickle  # type: ignore  # no stubs available
+import pydantic
 from griffe import Docstring, DocstringSectionKind, Parser, parse
-from typing_extensions import Literal
+from typing_extensions import Literal, TypeVar
 
+from prefect._internal.pydantic.v1_schema import has_v1_type_as_param
+from prefect._internal.pydantic.v2_schema import (
+    create_v2_schema,
+    process_v2_params,
+)
 from prefect.exceptions import (
+    MappingLengthMismatch,
+    MappingMissingIterable,
     ParameterBindError,
     ReservedArgumentError,
     SignatureMismatchError,
 )
 from prefect.logging.loggers import disable_logger, get_logger
+from prefect.utilities.annotations import allow_failure, quote, unmapped
+from prefect.utilities.collections import isiterable
 from prefect.utilities.importtools import safe_load_namespace
 
-logger = get_logger(__name__)
+logger: Logger = get_logger(__name__)
+
+R = TypeVar("R", infer_variance=True)
 
 
 def get_call_parameters(
-    fn: Callable,
-    call_args: Tuple[Any, ...],
-    call_kwargs: Dict[str, Any],
+    fn: Callable[..., Any],
+    call_args: tuple[Any, ...],
+    call_kwargs: dict[str, Any],
     apply_defaults: bool = True,
-) -> Dict[str, Any]:
+) -> dict[str, Any]:
     """
-    Bind a call to a function to get parameter/value mapping. Default values on the
-    signature will be included if not overridden.
+    Bind a call to a function to get parameter/value mapping. Default values on
+    the signature will be included if not overridden.
 
-    Raises a ParameterBindError if the arguments/kwargs are not valid for the function
+    If the function has a `__prefect_self__` attribute, it will be included as
+    the first parameter. This attribute is set when Prefect decorates a bound
+    method, so this approach allows Prefect to work with bound methods in a way
+    that is consistent with how Python handles them (i.e. users don't have to
+    pass the instance argument to the method) while still making the implicit self
+    argument visible to all of Prefect's parameter machinery (such as cache key
+    functions).
+
+    Raises a ParameterBindError if the arguments/kwargs are not valid for the
+    function
     """
+    if hasattr(fn, "__prefect_self__"):
+        call_args = (getattr(fn, "__prefect_self__"), *call_args)
+    elif hasattr(fn, "__prefect_cls__"):
+        call_args = (getattr(fn, "__prefect_cls__"), *call_args)
+
     try:
         bound_signature = inspect.signature(fn).bind(*call_args, **call_kwargs)
     except TypeError as exc:
@@ -65,14 +81,14 @@ def get_call_parameters(
 
 
 def get_parameter_defaults(
-    fn: Callable,
-) -> Dict[str, Any]:
+    fn: Callable[..., Any],
+) -> dict[str, Any]:
     """
     Get default parameter values for a callable.
     """
     signature = inspect.signature(fn)
 
-    parameter_defaults = {}
+    parameter_defaults: dict[str, Any] = {}
 
     for name, param in signature.parameters.items():
         if param.default is not signature.empty:
@@ -82,8 +98,8 @@ def get_parameter_defaults(
 
 
 def explode_variadic_parameter(
-    fn: Callable, parameters: Dict[str, Any]
-) -> Dict[str, Any]:
+    fn: Callable[..., Any], parameters: dict[str, Any]
+) -> dict[str, Any]:
     """
     Given a parameter dictionary, move any parameters stored in a variadic keyword
     argument parameter (i.e. **kwargs) into the top level.
@@ -116,8 +132,8 @@ def explode_variadic_parameter(
 
 
 def collapse_variadic_parameters(
-    fn: Callable, parameters: Dict[str, Any]
-) -> Dict[str, Any]:
+    fn: Callable[..., Any], parameters: dict[str, Any]
+) -> dict[str, Any]:
     """
     Given a parameter dictionary, move any parameters stored not present in the
     signature into the variadic keyword argument.
@@ -142,50 +158,47 @@ def collapse_variadic_parameters(
 
     missing_parameters = set(parameters.keys()) - set(signature_parameters.keys())
 
-    if not variadic_key and missing_parameters:
+    if not missing_parameters:
+        # no missing parameters, return parameters unchanged
+        return parameters
+
+    if not variadic_key:
         raise ValueError(
             f"Signature for {fn} does not include any variadic keyword argument "
             "but parameters were given that are not present in the signature."
         )
 
-    if variadic_key and not missing_parameters:
-        # variadic key is present but no missing parameters, return parameters unchanged
-        return parameters
-
     new_parameters = parameters.copy()
-    if variadic_key:
-        new_parameters[variadic_key] = {}
-
-    for key in missing_parameters:
-        new_parameters[variadic_key][key] = new_parameters.pop(key)
-
+    new_parameters[variadic_key] = {
+        key: new_parameters.pop(key) for key in missing_parameters
+    }
     return new_parameters
 
 
 def parameters_to_args_kwargs(
-    fn: Callable,
-    parameters: Dict[str, Any],
-) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
+    fn: Callable[..., Any],
+    parameters: dict[str, Any],
+) -> tuple[tuple[Any, ...], dict[str, Any]]:
     """
     Convert a `parameters` dictionary to positional and keyword arguments
 
     The function _must_ have an identical signature to the original function or this
     will return an empty tuple and dict.
     """
-    function_params = dict(inspect.signature(fn).parameters).keys()
+    function_params = inspect.signature(fn).parameters.keys()
     # Check for parameters that are not present in the function signature
     unknown_params = parameters.keys() - function_params
     if unknown_params:
         raise SignatureMismatchError.from_bad_params(
-            list(function_params), list(parameters.keys())
+            list(function_params), list(parameters)
         )
     bound_signature = inspect.signature(fn).bind_partial()
-    bound_signature.arguments = parameters
+    bound_signature.arguments = OrderedDict(parameters)
 
     return bound_signature.args, bound_signature.kwargs
 
 
-def call_with_parameters(fn: Callable, parameters: Dict[str, Any]):
+def call_with_parameters(fn: Callable[..., R], parameters: dict[str, Any]) -> R:
     """
     Call a function with parameters extracted with `get_call_parameters`
 
@@ -198,7 +211,7 @@ def call_with_parameters(fn: Callable, parameters: Dict[str, Any]):
 
 
 def cloudpickle_wrapped_call(
-    __fn: Callable, *args: Any, **kwargs: Any
+    __fn: Callable[..., Any], *args: Any, **kwargs: Any
 ) -> Callable[[], bytes]:
     """
     Serializes a function call using cloudpickle then returns a callable which will
@@ -208,18 +221,18 @@ def cloudpickle_wrapped_call(
     built-in pickler (e.g. `anyio.to_process` and `multiprocessing`) but may require
     a wider range of pickling support.
     """
-    payload = cloudpickle.dumps((__fn, args, kwargs))
+    payload = cloudpickle.dumps((__fn, args, kwargs))  # type: ignore   # no stubs available
     return partial(_run_serialized_call, payload)
 
 
-def _run_serialized_call(payload) -> bytes:
+def _run_serialized_call(payload: bytes) -> bytes:
     """
     Defined at the top-level so it can be pickled by the Python pickler.
     Used by `cloudpickle_wrapped_call`.
     """
     fn, args, kwargs = cloudpickle.loads(payload)
     retval = fn(*args, **kwargs)
-    return cloudpickle.dumps(retval)
+    return cloudpickle.dumps(retval)  # type: ignore   # no stubs available
 
 
 class ParameterSchema(pydantic.BaseModel):
@@ -227,19 +240,18 @@ class ParameterSchema(pydantic.BaseModel):
 
     title: Literal["Parameters"] = "Parameters"
     type: Literal["object"] = "object"
-    properties: Dict[str, Any] = pydantic.Field(default_factory=dict)
-    required: List[str] = None
-    definitions: Optional[Dict[str, Any]] = None
+    properties: dict[str, Any] = pydantic.Field(default_factory=dict)
+    required: list[str] = pydantic.Field(default_factory=list)
+    definitions: dict[str, Any] = pydantic.Field(default_factory=dict)
 
-    def dict(self, *args, **kwargs):
-        """Exclude `None` fields by default to comply with
-        the OpenAPI spec.
-        """
-        kwargs.setdefault("exclude_none", True)
-        return super().dict(*args, **kwargs)
+    def model_dump_for_openapi(self) -> dict[str, Any]:
+        result = self.model_dump(mode="python", exclude_none=True)
+        if "required" in result and not result["required"]:
+            del result["required"]
+        return result
 
 
-def parameter_docstrings(docstring: Optional[str]) -> Dict[str, str]:
+def parameter_docstrings(docstring: Optional[str]) -> dict[str, str]:
     """
     Given a docstring in Google docstring format, parse the parameter section
     and return a dictionary that maps parameter names to docstring.
@@ -250,7 +262,7 @@ def parameter_docstrings(docstring: Optional[str]) -> Dict[str, str]:
     Returns:
         Mapping from parameter names to docstrings.
     """
-    param_docstrings = {}
+    param_docstrings: dict[str, str] = {}
 
     if not docstring:
         return param_docstrings
@@ -271,35 +283,52 @@ def process_v1_params(
     param: inspect.Parameter,
     *,
     position: int,
-    docstrings: Dict[str, str],
-    aliases: Dict,
-) -> Tuple[str, Any, "pydantic.Field"]:
+    docstrings: dict[str, str],
+    aliases: dict[str, str],
+) -> tuple[str, Any, Any]:
+    import pydantic.v1 as pydantic_v1
+
     # Pydantic model creation will fail if names collide with the BaseModel type
-    if hasattr(pydantic.BaseModel, param.name):
+    if hasattr(pydantic_v1.BaseModel, param.name):
         name = param.name + "__"
         aliases[name] = param.name
     else:
         name = param.name
 
-    type_ = Any if param.annotation is inspect._empty else param.annotation
-    field = pydantic.Field(
-        default=... if param.default is param.empty else param.default,
-        title=param.name,
-        description=docstrings.get(param.name, None),
-        alias=aliases.get(name),
-        position=position,
-    )
+    type_ = Any if param.annotation is inspect.Parameter.empty else param.annotation
+
+    with warnings.catch_warnings():
+        # Note: pydantic.v1 doesn't have the warnings module, so we can't suppress them
+        warnings.filterwarnings("ignore", category=DeprecationWarning)
+        field: Any = pydantic_v1.Field(
+            default=... if param.default is param.empty else param.default,
+            title=param.name,
+            description=docstrings.get(param.name, None),
+            alias=aliases.get(name),
+            position=position,
+        )
     return name, type_, field
 
 
-def create_v1_schema(name_: str, model_cfg, **model_fields):
-    model: "pydantic.BaseModel" = pydantic.create_model(
-        name_, __config__=model_cfg, **model_fields
-    )
-    return model.schema(by_alias=True)
+def create_v1_schema(
+    name_: str, model_cfg: type[Any], model_fields: Optional[dict[str, Any]] = None
+) -> dict[str, Any]:
+    import pydantic.v1 as pydantic_v1
+
+    with warnings.catch_warnings():
+        # Note: pydantic.v1 doesn't have the warnings module, so we can't suppress them
+        warnings.filterwarnings("ignore", category=DeprecationWarning)
+
+        model_fields = model_fields or {}
+        model: type[pydantic_v1.BaseModel] = pydantic_v1.create_model(
+            name_,
+            __config__=model_cfg,
+            **model_fields,
+        )
+        return model.schema(by_alias=True)
 
 
-def parameter_schema(fn: Callable) -> ParameterSchema:
+def parameter_schema(fn: Callable[..., Any]) -> ParameterSchema:
     """Given a function, generates an OpenAPI-compatible description
     of the function's arguments, including:
         - name
@@ -360,7 +389,7 @@ def parameter_schema_from_entrypoint(entrypoint: str) -> ParameterSchema:
 
 
 def generate_parameter_schema(
-    signature: inspect.Signature, docstrings: Dict[str, str]
+    signature: inspect.Signature, docstrings: dict[str, str]
 ) -> ParameterSchema:
     """
     Generate a parameter schema from a function signature and docstrings.
@@ -376,47 +405,54 @@ def generate_parameter_schema(
         ParameterSchema: The parameter schema.
     """
 
-    model_fields = {}
-    aliases = {}
+    model_fields: dict[str, Any] = {}
+    aliases: dict[str, str] = {}
 
-    if HAS_PYDANTIC_V2 and not has_v1_type_as_param(signature):
-        create_schema = create_v2_schema
+    if not has_v1_type_as_param(signature):
+        config = pydantic.ConfigDict(arbitrary_types_allowed=True)
+
+        create_schema = partial(create_v2_schema, model_cfg=config)
         process_params = process_v2_params
-        ModelConfig = pydantic.ConfigDict(arbitrary_types_allowed=True)
+
     else:
-        create_schema = create_v1_schema
-        process_params = process_v1_params
 
         class ModelConfig:
             arbitrary_types_allowed = True
 
+        create_schema = partial(create_v1_schema, model_cfg=ModelConfig)
+        process_params = process_v1_params
+
     for position, param in enumerate(signature.parameters.values()):
         name, type_, field = process_params(
             param, position=position, docstrings=docstrings, aliases=aliases
         )
+
+        if name in ("cls", "self"):
+            continue  # Exclude 'cls'/'self' as they're implicitly passed and not real flow parameters
+
         # Generate a Pydantic model at each step so we can check if this parameter
         # type supports schema generation
         try:
-            create_schema(
-                "CheckParameter", model_cfg=ModelConfig, **{name: (type_, field)}
-            )
+            create_schema("CheckParameter", model_fields={name: (type_, field)})
         except (ValueError, TypeError):
             # This field's type is not valid for schema creation, update it to `Any`
             type_ = Any
         model_fields[name] = (type_, field)
 
     # Generate the final model and schema
-    schema = create_schema("Parameters", model_cfg=ModelConfig, **model_fields)
+    schema = create_schema("Parameters", model_fields=model_fields)
     return ParameterSchema(**schema)
 
 
-def raise_for_reserved_arguments(fn: Callable, reserved_arguments: Iterable[str]):
+def raise_for_reserved_arguments(
+    fn: Callable[..., Any], reserved_arguments: Iterable[str]
+) -> None:
     """Raise a ReservedArgumentError if `fn` has any parameters that conflict
     with the names contained in `reserved_arguments`."""
-    function_paremeters = inspect.signature(fn).parameters
+    function_parameters = inspect.signature(fn).parameters
 
     for argument in reserved_arguments:
-        if argument in function_paremeters:
+        if argument in function_parameters:
             raise ReservedArgumentError(
                 f"{argument!r} is a reserved argument name and cannot be used."
             )
@@ -460,7 +496,7 @@ def _generate_signature_from_source(
     )
     if func_def is None:
         raise ValueError(f"Function {func_name} not found in source code")
-    parameters = []
+    parameters: list[inspect.Parameter] = []
 
     # Handle annotations for positional only args e.g. def func(a, /, b, c)
     for arg in func_def.args.posonlyargs:
@@ -618,5 +654,77 @@ def _get_docstring_from_source(source_code: str, func_name: str) -> Optional[str
         and isinstance(func_def.body[0], ast.Expr)
         and isinstance(func_def.body[0].value, ast.Constant)
     ):
-        return func_def.body[0].value.value
+        return str(func_def.body[0].value.value)
     return None
+
+
+def expand_mapping_parameters(
+    func: Callable[..., Any], parameters: dict[str, Any]
+) -> list[dict[str, Any]]:
+    """
+    Generates a list of call parameters to be used for individual calls in a mapping
+    operation.
+
+    Args:
+        func: The function to be called
+        parameters: A dictionary of parameters with iterables to be mapped over
+
+    Returns:
+        list: A list of dictionaries to be used as parameters for each
+            call in the mapping operation
+    """
+    # Ensure that any parameters in kwargs are expanded before this check
+    parameters = explode_variadic_parameter(func, parameters)
+
+    iterable_parameters: dict[str, list[Any]] = {}
+    static_parameters: dict[str, Any] = {}
+    annotated_parameters: dict[str, Union[allow_failure[Any], quote[Any]]] = {}
+    for key, val in parameters.items():
+        if isinstance(val, (allow_failure, quote)):
+            # Unwrap annotated parameters to determine if they are iterable
+            annotated_parameters[key] = val
+            val: Any = val.unwrap()
+
+        if isinstance(val, unmapped):
+            static_parameters[key] = cast(unmapped[Any], val).value
+        elif isiterable(val):
+            iterable_parameters[key] = list(val)
+        else:
+            static_parameters[key] = val
+
+    if not iterable_parameters:
+        raise MappingMissingIterable(
+            "No iterable parameters were received. Parameters for map must "
+            f"include at least one iterable. Parameters: {parameters}"
+        )
+
+    iterable_parameter_lengths = {
+        key: len(val) for key, val in iterable_parameters.items()
+    }
+    lengths = set(iterable_parameter_lengths.values())
+    if len(lengths) > 1:
+        raise MappingLengthMismatch(
+            "Received iterable parameters with different lengths. Parameters for map"
+            f" must all be the same length. Got lengths: {iterable_parameter_lengths}"
+        )
+
+    map_length = list(lengths)[0]
+
+    call_parameters_list: list[dict[str, Any]] = []
+    for i in range(map_length):
+        call_parameters = {key: value[i] for key, value in iterable_parameters.items()}
+        call_parameters.update({key: value for key, value in static_parameters.items()})
+
+        # Add default values for parameters; these are skipped earlier since they should
+        # not be mapped over
+        for key, value in get_parameter_defaults(func).items():
+            call_parameters.setdefault(key, value)
+
+        # Re-apply annotations to each key again
+        for key, annotation in annotated_parameters.items():
+            call_parameters[key] = annotation.rewrap(call_parameters[key])
+
+        # Collapse any previously exploded kwargs
+        call_parameters_list.append(collapse_variadic_parameters(func, call_parameters))
+
+    return call_parameters_list
